package com.hartmath.util;

import com.hartmath.expression.*;
import com.hartmath.mapping.*;
import com.hartmath.lib.*;

/**
* The ComplexMatrix class provides an object for encapsulating matrices containing complex numbers.
* @version 2.0
* @author Mark Hale
*/
public class ComplexMatrix extends Matrix {
        /**
        * Array containing the elements of the matrix.
        */
        protected HDoubleComplex matrix[][];
        /**
        * Constructs a matrix.
        */
        protected ComplexMatrix() {}
        /**
        * Constructs an empty matrix.
        * @param row the number of rows
        * @param col the number of columns
        */
        public ComplexMatrix(final int row,final int col) {
                matrix=new HDoubleComplex[row][col];
        }
        /**
        * Constructs a matrix by wrapping an array.
        * @param array an assigned value
        */
        public ComplexMatrix(final HDoubleComplex array[][]) {
                matrix=array;
        }
        /**
        * Constructs a matrix from an array of vectors (columns).
        * @param array an assigned value
        */
        public ComplexMatrix(ComplexVector array[]) {
                this(array.length,array[0].dimension());
                for(int j,i=0;i<matrix[0].length;i++) {
                        for(j=0;j<matrix.length;j++)
                                matrix[i][j]=array[j].getComponent(i);
                }
        }
        /**
        * Finalize.
        * @exception Throwable Any that occur.
        */
        protected void finalize() throws Throwable {
                matrix=null;
                super.finalize();
        }
        /**
        * Compares two complex matrices for equality.
        * @param m a complex matrix
        */
        public boolean equals(Object m) {
                if(m!=null && (m instanceof ComplexMatrix) &&
                matrix.length==((ComplexMatrix)m).rows() && matrix[0].length==((ComplexMatrix)m).columns()) {
                        final ComplexMatrix cm=(ComplexMatrix)m;
                        for(int j,i=0;i<matrix.length;i++) {
                                for(j=0;j<matrix[0].length;j++) {
                                        if(!matrix[i][j].equals(cm.getElement(i,j)))
                                                return false;
                                }
                        }
                        return true;
                } else {
                        return false;
                }
        }
        /**
        * Returns a string representing this matrix.
        */
        public String toString() {
                final StringBuffer buf=new StringBuffer(matrix.length*matrix[0].length);
                for(int j,i=0;i<matrix.length;i++) {
                        for(j=0;j<matrix[0].length;j++) {
                                buf.append(matrix[i][j].toString());
                                buf.append(' ');
                        }
                        buf.append('\n');
                }
                return buf.toString();
        }
        public HFunction toFunction() {
                HFunction mat = new HFunction( C.List );
                mat.setProperties(HFunction.MATRIXFORM);
                for(int i=0;i<matrix.length;i++) {
                  mat.add(new HFunction( C.List, matrix[i] ) );
                }
                return mat;
        }
        
        /**
        * Returns a hashcode for this matrix.
        */
        public int hashCode() {
                return (int)Math.exp(infNorm());
        }
        /**
        * Returns the real part of this complex matrix.
        * @return a double matrix
        */
        public DoubleMatrix real() {
                final double array[][]=new double[matrix.length][matrix[0].length];
                for(int j,i=0;i<matrix.length;i++) {
                        for(j=0;j<matrix[0].length;j++)
                                array[i][j]=matrix[i][j].real();
                }
                return new DoubleMatrix(array);
        }
        /**
        * Returns the imaginary part of this complex matrix.
        * @return a double matrix
        */
        public DoubleMatrix imag() {
                final double array[][]=new double[matrix.length][matrix[0].length];
                for(int j,i=0;i<matrix.length;i++) {
                        for(j=0;j<matrix[0].length;j++)
                                array[i][j]=matrix[i][j].imag();
                }
                return new DoubleMatrix(array);
        }
        /**
        * Returns an element of the matrix.
        * @param i row index of the element
        * @param j column index of the element
        * @exception MatrixDimensionException If attempting to access an invalid element.
        */
        public HDoubleComplex getElement(final int i, final int j) {
                if(i>=0 && i<matrix.length && j>=0 && j<matrix[0].length)
                        return matrix[i][j];
                else
              //          throw new MatrixDimensionException("Invalid element.");
                  throw new HThrowException(C.OutOfRangeError, C.List,
                                new HString("Invalid element."));
        }
        /**
        * Sets the value of an element of the matrix.
        * @param i row index of the element
        * @param j column index of the element
        * @param z a complex number
        * @exception MatrixDimensionException If attempting to access an invalid element.
        */
        public void setElement(final int i, final int j, final HDoubleComplex z) {
                if(i>=0 && i<matrix.length && j>=0 && j<matrix[0].length)
                        matrix[i][j]=z;
                else
                   //     throw new MatrixDimensionException("Invalid element.");
                  throw new HThrowException(C.OutOfRangeError, C.List,
                                new HString("Invalid element."));
        }
        /**
        * Returns the l(infinity)-norm.
        * @author Taber Smith
        */
        public double infNorm() {
                double result=0.0,tmpResult;
                for(int j,i=0;i<matrix.length;i++) {
                        tmpResult=0.0;
                        for(j=0;j<matrix[0].length;j++)
                                tmpResult+=matrix[i][j].mod();
                        if(tmpResult>result)
                                result=tmpResult;
                }
                return result;
        }
        /**
        * Returns the Frobenius norm.
        * @author Taber Smith
        */
        public double frobeniusNorm() {
                double result=0.0;
                for(int j,i=0;i<matrix.length;i++)
                        for(j=0;j<matrix[0].length;j++)
                                result+=matrix[i][j].mod()*matrix[i][j].mod();
                return Math.sqrt(result);
        }
        /**
        * Returns the number of rows.
        */
        public int rows() {
                return matrix.length;
        }
        /**
        * Returns the number of columns.
        */
        public int columns() {
                return matrix[0].length;
        }

//============
// OPERATIONS
//============

// ADDITION

        /**
        * Returns the addition of this matrix and another.
        */
        public Matrix add(final Matrix m) {
                if(m instanceof ComplexMatrix)
                        return add((ComplexMatrix)m);
                else
                        throw new IllegalArgumentException("Matrix class not recognised by this method.");
        }
        /**
        * Returns the addition of this matrix and another.
        * @param m a complex matrix
        * @exception MatrixDimensionException If the matrices are different sizes.
        */
        public ComplexMatrix add(final ComplexMatrix m) {
                if(matrix.length==m.rows() && matrix[0].length==m.columns()) {
                        final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][matrix[0].length];
                        for(int j,i=0;i<array.length;i++) {
                                array[i][0]=matrix[i][0].add(m.getElement(i,0));
                                for(j=1;j<array[0].length;j++)
                                        array[i][j]=matrix[i][j].add(m.getElement(i,j));
                        }
                        return new ComplexMatrix(array);
                } else
              //          throw new MatrixDimensionException("Matrices are different sizes.");
                  throw new HThrowException(C.OutOfRangeError, C.Add,
                                new HString("Matrices have different sizes."));
        }

// SUBTRACTION

        /**
        * Returns the subtraction of this matrix by another.
        */
        public Matrix subtract(final Matrix m) {
                if(m instanceof ComplexMatrix)
                        return subtract((ComplexMatrix)m);
                else
                        throw new IllegalArgumentException("Matrix class not recognised by this method.");
        }
        /**
        * Returns the subtraction of this matrix by another.
        * @param m a complex matrix
        * @exception MatrixDimensionException If the matrices are different sizes.
        */
        public ComplexMatrix subtract(final ComplexMatrix m) {
                if(matrix.length==m.rows() && matrix[0].length==m.columns()) {
                        final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][matrix[0].length];
                        for(int j,i=0;i<array.length;i++) {
                                array[i][0]=matrix[i][0].subtract(m.getElement(i,0));
                                for(j=1;j<array[0].length;j++)
                                        array[i][j]=matrix[i][j].subtract(m.getElement(i,j));
                        }
                        return new ComplexMatrix(array);
                } else
                   //     throw new MatrixDimensionException("Matrices are different sizes.");
                  throw new HThrowException(C.OutOfRangeError, C.Subtract,
                                new HString("Matrices have different sizes."));
        }

// SCALAR MULTIPLICATION

        /**
        * Returns the multiplication of this matrix by a scalar.
        * @param z a complex number
        * @return a complex matrix
        */
        public ComplexMatrix scalarMultiply(final HDoubleComplex z) {
                final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][matrix[0].length];
                for(int j,i=0;i<array.length;i++) {
                        array[i][0]=z.multiply(matrix[i][0]);
                        for(j=1;j<array[0].length;j++)
                                array[i][j]=z.multiply(matrix[i][j]);
                }
                return new ComplexMatrix(array);
        }
        /**
        * Returns the multiplication of this matrix by a scalar.
        * @param x a double
        * @return a complex matrix
        */
        public ComplexMatrix scalarMultiply(final double x) {
                final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][matrix[0].length];
                for(int j,i=0;i<array.length;i++) {
                        array[i][0]=matrix[i][0].multiply(x);
                        for(j=1;j<array[0].length;j++)
                                array[i][j]=matrix[i][j].multiply(x);
                }
                return new ComplexMatrix(array);
        }

// MATRIX MULTIPLICATION

        /**
        * Returns the multiplication of a vector by this matrix.
        * @param v a complex vector
        * @exception DimensionException If the matrix and vector are incompatible.
        */
        public ComplexVector multiply(final ComplexVector v) {
                if(matrix[0].length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[matrix.length];
                        for(int j,i=0;i<array.length;i++) {
                                array[i]=matrix[i][0].multiply(v.getComponent(0));
                                for(j=1;j<matrix[0].length;j++)
                                        array[i]=array[i].add(matrix[i][j].multiply(v.getComponent(j)));
                        }
                        return new ComplexVector(array);
                } else
                //        throw new DimensionException("Matrix and vector are incompatible.");
                  throw new HThrowException(C.OutOfRangeError, C.Add,
                                new HString("Matrix and vector are incompatible."));
        }
        /**
        * Returns the multiplication of this matrix and another.
        */
        public Matrix multiply(final Matrix m) {
                if(m instanceof ComplexMatrix)
                        return multiply((ComplexMatrix)m);
                else
                        throw new IllegalArgumentException("Matrix class not recognised by this method.");
        }
        /**
        * Returns the multiplication of this matrix and another.
        * @param m a complex matrix
        * @return a ComplexMatrix or a casted ComplexSquareMatrix as appropriate
        * @exception MatrixDimensionException If the matrices are incompatible.
        */
        public ComplexMatrix multiply(final ComplexMatrix m) {
                if(matrix[0].length==m.rows()) {
                        int n,k;
                        final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][m.columns()];
                        for(int j=0;j<array.length;j++) {
                                for(k=0;k<array[0].length;k++) {
                                        array[j][k]=matrix[j][0].multiply(m.getElement(0,k));
                                        for(n=1;n<matrix[0].length;n++)
                                                array[j][k]=array[j][k].add(matrix[j][n].multiply(m.getElement(n,k)));
                                }
                        }
                      //  if(array.length==array[0].length)
                      //          return new ComplexSquareMatrix(array);
                      //  else
                                return new ComplexMatrix(array);
                } else
                //        throw new MatrixDimensionException("Incompatible matrices.");
                  throw new HThrowException(C.OutOfRangeError, C.Multiply,
                                new HString("Incompatible matrices."));

        }


// HERMITIAN ADJOINT

        /**
        * Returns the hermitian adjoint of this matrix.
        * @return a complex matrix
        */
        public ComplexMatrix hermitianAdjoint() {
                final HDoubleComplex array[][]=new HDoubleComplex[matrix[0].length][matrix.length];
                for(int j,i=0;i<array.length;i++) {
                        array[0][i]=matrix[i][0].conj();
                        for(j=1;j<array[0].length;j++)
                                array[j][i]=matrix[i][j].conj();
                }
                return new ComplexMatrix(array);
        }

// CONJUGATE

        /**
        * Returns the complex conjugate of this matrix.
        * @return a complex matrix
        */
        public ComplexMatrix conjugate() {
                final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][matrix[0].length];
                for(int j,i=0;i<array.length;i++) {
                        array[i][0]=matrix[i][0].conj();
                        for(j=1;j<array[0].length;j++)
                                array[i][j]=matrix[i][j].conj();
                }
                return new ComplexMatrix(array);
        }

// TRANSPOSE

        /**
        * Returns the transpose of this matrix.
        * @return a complex matrix
        */
        public Matrix transpose() {
                final HDoubleComplex array[][]=new HDoubleComplex[matrix[0].length][matrix.length];
                for(int j,i=0;i<array[0].length;i++) {
                        array[0][i]=matrix[i][0];
                        for(j=1;j<array.length;j++)
                                array[j][i]=matrix[i][j];
                }
                return new ComplexMatrix(array);
        }

// MAP ELEMENTS

        /**
        * Applies a function on all the matrix elements.
        * @param f a user-defined function
        * @return a complex matrix
        */
        public ComplexMatrix mapElements(final HUnaryNumerical f) {
                final HDoubleComplex array[][]=new HDoubleComplex[matrix.length][matrix[0].length];
                for(int j,i=0;i<array.length;i++) {
                        array[i][0]=f.map(matrix[i][0]);
                        for(j=1;j<array[0].length;j++)
                                array[i][j]=f.map(matrix[i][j]);
                }
                return new ComplexMatrix(array);
        }
}