package com.hartmath.util;

import com.hartmath.lib.*;
import com.hartmath.expression.*;
import com.hartmath.mapping.*;
//import com.objectspace.jgl.*;

/**
* The ComplexVector class encapsulates vectors containing complex numbers.
* @version 2.0
* @author Mark Hale
*/
public class ComplexVector extends MathVector {
        protected HDoubleComplex vector[];

        /**
        * Constructs an empty vector.
        * @param dim the dimension of the vector.
        */
        public ComplexVector(final int dim) {
                vector=new HDoubleComplex[dim];
        }
        /**
        * Constructs a vector by wrapping an array.
        * @param array an assigned value
        */
        public ComplexVector(final HDoubleComplex array[]) {
                vector=array;
        }
        /**
        * Compares two complex vectors for equality.
        * @param a a complex vector
        */
        public boolean equals(Object a) {
                if(a!=null && (a instanceof ComplexVector) && vector.length==((ComplexVector)a).dimension()) {
                        final ComplexVector cv=(ComplexVector)a;
                        for(int i=0;i<vector.length;i++) {
                                if(!vector[i].equals(cv.getComponent(i)))
                                        return false;
                        }
                        return true;
                } else
                        return false;
        }
        /**
        * Returns a comma delimited string representing the value of this vector.
        */
        public String toString() {
                final StringBuffer buf=new StringBuffer(vector.length);
                int i;
                for(i=0;i<vector.length-1;i++) {
                        buf.append(vector[i].toString());
                        buf.append(',');
                }
                buf.append(vector[i].toString());
                return buf.toString();
        }
        public HFunction toFunction() {
                return new HFunction(C.List, vector );
        }
        /**
        * Returns a hashcode for this vector.
        */
        public int hashCode() {
                return (int)Math.exp(norm());
        }
        /**
        * Returns the real part of this complex vector.
        */
        public DoubleVector real() {
                final double array[]=new double[vector.length];
                for(int i=0;i<vector.length;i++)
                        array[i]=vector[i].real();
                return new DoubleVector(array);
        }
        /**
        * Returns the imaginary part of this complex vector.
        */
        public DoubleVector imag() {
                final double array[]=new double[vector.length];
                for(int i=0;i<vector.length;i++)
                        array[i]=vector[i].imag();
                return new DoubleVector(array);
        }
        /**
        * Returns a component of this vector.
        * @param n index of the vector component
        * @exception VectorDimensionException If attempting to access an invalid component.
        */
        public HDoubleComplex getComponent(final int n) {
                if(n>=0 && n<vector.length)
                        return vector[n];
                else
         //               throw new VectorDimensionException("Invalid component.");
                  throw new HThrowException(C.OutOfRangeError, C.List, 
                                new HString("Invalid component."));
        }
        /**
        * Sets the value of a component of this vector.
        * @param n index of the vector component
        * @param z a complex number
        * @exception VectorDimensionException If attempting to access an invalid component.
        */
        public void setComponent(final int n, final HDoubleComplex z) {
                if(n>=0 && n<vector.length)
                        vector[n]=z;
                else
               //         throw new VectorDimensionException("Invalid component.");
                  throw new HThrowException(C.OutOfRangeError, C.List, 
                                new HString("Invalid component."));
        }
        /**
        * Returns the vector's dimension.
        */
        public int dimension() {
                return vector.length;
        }
        /**
        * Returns the l2-norm (magnitude).
        */
        public double norm() {
                HDoubleComplex answer=vector[0].multiply(vector[0].conj());
                for(int i=1;i<vector.length;i++)
                        answer=answer.add(vector[i].multiply(vector[i].conj()));
                return Math.sqrt(answer.real());
        }
        /**
        * Returns the l(infinity)-norm.
        * @author Taber Smith
        */
        public double infNorm() {
                double infNorm=vector[0].mod();
                for(int i=1;i<vector.length;i++) {
                        if(vector[i].mod()>infNorm)
                                infNorm=vector[i].mod();
                }
                return infNorm;
        }

//============
// OPERATIONS
//============

// COMPLEX CONJUGATE

        /**
        * Returns the complex conjugate of this vector.
        */
        public ComplexVector conjugate() {
                final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                array[0]=vector[0].conj();
                for(int i=1;i<array.length;i++)
                        array[i]=vector[i].conj();
                return new ComplexVector(array);
        }

// ADDITION

        /**
        * Returns the addition of this vector and another.
        */
        public MathVector add(final MathVector v) {
                if(v instanceof ComplexVector)
                        return add((ComplexVector)v);
                else if(v instanceof DoubleVector)
                        return add((DoubleVector)v);
         //       else if(v instanceof IntegerVector)
         //               return add((IntegerVector)v);
                else
                        throw new IllegalArgumentException("Vector class not recognised by this method.");
        }
        /**
        * Returns the addition of this vector and another.
        * @param v a complex vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */
        public ComplexVector add(final ComplexVector v) {
                if(vector.length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                        array[0]=vector[0].add(v.getComponent(0));
                        for(int i=1;i<array.length;i++)
                                array[i]=vector[i].add(v.getComponent(i));
                        return new ComplexVector(array);
                } else
                 //   throw new VectorDimensionException("Vectors are different sizes.");
                 throw new HThrowException(C.OutOfRangeError, C.Add,
                                new HString("ComplexVectors have different sizes."));
        }
        /**
        * Returns the addition of this vector and another.
        * @param v a double vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */
        public ComplexVector add(final DoubleVector v) {
                if(vector.length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                        array[0]=vector[0].addReal(v.getComponent(0));
                        for(int i=1;i<array.length;i++)
                                array[i]=vector[i].addReal(v.getComponent(i));
                        return new ComplexVector(array);
                } else
                 //       throw new VectorDimensionException("Vectors are different sizes.");
                  throw new HThrowException(C.OutOfRangeError, C.Add,
                                new HString("Vectors have different sizes."));

        }
        /**
        * Returns the addition of this vector and another.
        * @param v an integer vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */    /*
        public ComplexVector add(final IntegerVector v) {
                if(vector.length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                        array[0]=vector[0].addReal(v.getComponent(0));
                        for(int i=1;i<array.length;i++)
                                array[i]=vector[i].addReal(v.getComponent(i));
                        return new ComplexVector(array);
                } else
                        throw new VectorDimensionException("Vectors are different sizes.");
        }       */

// SUBTRACTION

        /**
        * Returns the subtraction of this vector by another.
        */
        public MathVector subtract(final MathVector v) {
                if(v instanceof ComplexVector)
                        return subtract((ComplexVector)v);
                else if(v instanceof DoubleVector)
                        return subtract((DoubleVector)v);
         //       else if(v instanceof IntegerVector)
         //               return subtract((IntegerVector)v);
                else
                        throw new IllegalArgumentException("Vector class not recognised by this method.");
        }
        /**
        * Returns the subtraction of this vector by another.
        * @param v a complex vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */
        public ComplexVector subtract(final ComplexVector v) {
                if(vector.length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                        array[0]=vector[0].subtract(v.getComponent(0));
                        for(int i=1;i<array.length;i++)
                                array[i]=vector[i].subtract(v.getComponent(i));
                        return new ComplexVector(array);
                } else
                 //       throw new VectorDimensionException("Vectors are different sizes.");
                   throw new HThrowException(C.OutOfRangeError, C.Subtract,
                                new HString("Vectors have different sizes."));
        }
        /**
        * Returns the subtraction of this vector by another.
        * @param v a double vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */
        public ComplexVector subtract(final DoubleVector v) {
                if(vector.length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                        array[0]=vector[0].subtractReal(v.getComponent(0));
                        for(int i=1;i<array.length;i++)
                                array[i]=vector[i].subtractReal(v.getComponent(i));
                        return new ComplexVector(array);
                } else
              //        throw new VectorDimensionException("Vectors are different sizes.");
                  throw new HThrowException(C.OutOfRangeError, C.Subtract,
                                new HString("Vectors have different sizes."));
        }
        /**
        * Returns the subtraction of this vector by another.
        * @param v an integer vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */       /*
        public ComplexVector subtract(final IntegerVector v) {
                if(vector.length==v.dimension()) {
                        final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                        array[0]=vector[0].subtractReal(v.getComponent(0));
                        for(int i=1;i<array.length;i++)
                                array[i]=vector[i].subtractReal(v.getComponent(i));
                        return new ComplexVector(array);
                } else
                        throw new VectorDimensionException("Vectors are different sizes.");
        }          */

// SCALAR MULTIPLICATION

        /**
        * Returns the multiplication of this vector by a scalar.
        */
        public MathVector scalarMultiply(MathNumber x) {
                if(x instanceof HDoubleComplex)
                        return scalarMultiply((HDoubleComplex)x);
                else if(x instanceof HDouble)
                        return scalarMultiply(new HDoubleComplex(((HDouble)x).value(),0.0));
        //        else if(x instanceof MathInteger)
        //                return scalarMultiply(new HDoubleComplex(((MathInteger)x).value(),0.0));
                else
                        throw new IllegalArgumentException("Number class not recognised by this method.");
        }
        /**
        * Returns the multiplication of this vector by a scalar.
        * @param z a complex number
        */
        public ComplexVector scalarMultiply(final HDoubleComplex z) {
                final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                array[0]=vector[0].multiply(z);
                for(int i=1;i<array.length;i++)
                        array[i]=vector[i].multiply(z);
                return new ComplexVector(array);
        }

// SCALAR DIVISION

        /**
        * Returns the division of this vector by a scalar.
        * @param z a complex number
        * @exception ArithmeticException If divide by zero.
        */
        public ComplexVector scalarDivide(final HDoubleComplex z) {
                final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                array[0]=vector[0].divide(z);
                for(int i=1;i<array.length;i++)
                        array[i]=vector[i].divide(z);
                return new ComplexVector(array);
        }

// SCALAR PRODUCT

        /**
        * Returns the scalar product of this vector and another.
        * @param v a complex vector
        * @exception VectorDimensionException If the vectors are different sizes.
        */
        public HDoubleComplex scalarProduct(final ComplexVector v) {
                if(vector.length==v.dimension()) {
                        HDoubleComplex answer=vector[0].multiply(v.getComponent(0).conj());
                        for(int i=1;i<vector.length;i++)
                                answer=answer.add(vector[i].multiply(v.getComponent(i).conj()));
                        return answer;
                } else
               //         throw new VectorDimensionException("Vectors are different sizes.");
                    throw new HThrowException(C.OutOfRangeError, C.Multiply,
                                new HString("Vectors have different sizes."));

        }

// MAP COMPONENTS

        /**
        * Applies a function on all the vector components.
        * @param f a user-defined function
        * @return a complex vector
        */
        public ComplexVector mapComponents(final HUnaryNumerical f) {
                final HDoubleComplex array[]=new HDoubleComplex[vector.length];
                array[0]=f.map(vector[0]);
                for(int i=1;i<vector.length;i++)
                        array[i]=f.map(vector[i]);
                return new ComplexVector(array);
        }
}

