/*--- formatted by Jindent 2.0b2, (www.c-lab.de/~jindent) ---*/

package com.hartmath.classes;

import com.hartmath.lib.*;
import com.hartmath.util.*;
import com.hartmath.expression.*;
import com.hartmath.mapping.*;

/**
* The Fourier math library.
* This class cannot be subclassed or instantiated because all methods are static.
* @version 0.7
* @author Mark Hale
*/
final class FourierMath {

        /**
        * Fourier transform.
        * @return an array containing positive frequencies in ascending order
        * followed by negative frequencies in ascending order
        * @author Don Cross
        */
        public static HDoubleComplex[] transform(final HDoubleComplex data[]) {
                final double array[][]=new double[2][data.length];
                int i,j,k,n;
                int numBits,blockSize,blockEnd;
                double deltaAngle,angleRe,angleIm,tmpRe,tmpIm;
                double alpha,beta; // used in recurrence relation

        //        if(!isPowerOf2(data.length))
        //                throw new IllegalArgumentException("The number of samples must be a power of 2.");

                numBits=numberOfBitsNeeded(data.length);
// Simultaneous data copy and bit-reversal ordering into output
                for(i=0;i<data.length;i++) {
                        j=reverseBits(i,numBits);
                        array[0][j]=data[i].real();
                        array[1][j]=data[i].imag();
                }
// FFT
                blockEnd=1;
                for(blockSize=2;blockSize<=data.length;blockSize<<=1) {
                        deltaAngle=HDouble.TWO_PI/blockSize;
                        alpha=Math.sin(0.5*deltaAngle);
                        alpha=2.0*alpha*alpha;
                        beta=Math.sin(deltaAngle);
                        for(i=0;i<data.length;i+=blockSize) {
                                angleRe=1.0;angleIm=0.0;
                                for(j=i,n=0;n<blockEnd;j++,n++) {
                                        k=j+blockEnd;
                                        tmpRe=angleRe*array[0][k]-angleIm*array[1][k];
                                        tmpIm=angleRe*array[1][k]+angleIm*array[0][k];
                                        array[0][k]=array[0][j]-tmpRe;
                                        array[1][k]=array[1][j]-tmpIm;
                                        array[0][j]+=tmpRe;array[1][j]+=tmpIm;
                                        tmpRe=alpha*angleRe+beta*angleIm;
                                        tmpIm=alpha*angleIm-beta*angleRe;
                                        angleRe-=tmpRe;angleIm-=tmpIm;
                                }
                        }
                        blockEnd=blockSize;
                }
                final HDoubleComplex answer[]=new HDoubleComplex[data.length];
                for(i=0;i<data.length;i++)
                        answer[i]=new HDoubleComplex(array[0][i],array[1][i]);
                return answer;
        }
        
        /**
        * Inverse Fourier transform.
        * @return an array containing the positive time part of the signal
        * followed by the negative time part
        * @author Don Cross
        */
        public static HDoubleComplex[] inverseTransform(final HDoubleComplex data[]) {
                final double array[][]=new double[2][data.length];
                int i,j,k,n;
                int numBits,blockSize,blockEnd;
                double deltaAngle,angleRe,angleIm,tmpRe,tmpIm;
                double alpha,beta; // used in recurrence relation

           //     if(!isPowerOf2(data.length))
           //             throw new IllegalArgumentException("Data length must be a power of 2.");

                numBits=numberOfBitsNeeded(data.length);
// Simultaneous data copy and bit-reversal ordering into output
                for(i=0;i<data.length;i++) {
                        j=reverseBits(i,numBits);
                        array[0][j]=data[i].real();
                        array[1][j]=data[i].imag();
                }
// FFT
                blockEnd=1;
                for(blockSize=2;blockSize<=data.length;blockSize<<=1) {
                        deltaAngle=-HDouble.TWO_PI/blockSize;
                        alpha=Math.sin(0.5*deltaAngle);
                        alpha=2.0*alpha*alpha;
                        beta=Math.sin(deltaAngle);
                        for(i=0;i<data.length;i+=blockSize) {
                                angleRe=1.0;angleIm=0.0;
                                for(j=i,n=0;n<blockEnd;j++,n++) {
                                        k=j+blockEnd;
                                        tmpRe=angleRe*array[0][k]-angleIm*array[1][k];
                                        tmpIm=angleRe*array[1][k]+angleIm*array[0][k];
                                        array[0][k]=array[0][j]-tmpRe;
                                        array[1][k]=array[1][j]-tmpIm;
                                        array[0][j]+=tmpRe;array[1][j]+=tmpIm;
                                        tmpRe=alpha*angleRe+beta*angleIm;
                                        tmpIm=alpha*angleIm-beta*angleRe;
                                        angleRe-=tmpRe;angleIm-=tmpIm;
                                }
                        }
                        blockEnd=blockSize;
                }
// Normalize
                final HDoubleComplex answer[]=new HDoubleComplex[data.length];
                final double denom=data.length;
                for(i=0;i<data.length;i++)
                        answer[i]=new HDoubleComplex(array[0][i]/denom,array[1][i]/denom);
                return answer;
        }
        /**
        * Number of bits needed.
        * @author Don Cross
        */
        private static int numberOfBitsNeeded(final int pwrOf2) {
                if(pwrOf2<2)
                        throw new IllegalArgumentException();
                for(int i=0;;i++) {
                        if((pwrOf2&(1<<i))>0)
                                return i;
                }
        }
        /**
        * Reverse bits.
        * @author Don Cross
        */
        private static int reverseBits(int index,final int numBits) {
                int i,rev;
                for(i=rev=0;i<numBits;i++) {
                        rev=(rev<<1)|(index&1);
                        index>>=1;
                }
                return rev;
        }

        /**
        * Sorts the output from the Fourier transfom methods into
        * ascending frequency/time order.
        */
        public static HDoubleComplex[] sort(final HDoubleComplex output[]) {
                final HDoubleComplex ret[]=new HDoubleComplex[output.length];
                final int Nby2=output.length/2;
                for(int i=0;i<Nby2;i++) {
                        ret[Nby2+i]=output[i];
                        ret[i]=output[Nby2+i];
                }
                return ret;
        }
}


/**
 * Combinatoric algorithms
 * 
 * @author Klaus Hartlage  <A HREF="mailto:khartlage@t-online.de">khartlage@t-online.de</A>
 */
public class Fourier implements FunctionEvaluator {

	/**
	 * Class Declaration.
	 * 
	 * 
	 * @see
	 * 
	 * @author
	 * @version   %I%, %G%
	 */
	class EFFT implements FunctionEvaluator {

		/**
		 * Method Declaration.
		 * 
		 * 
		 * @param vect
		 * 
		 * @return
		 * 
		 * @see
		 */
		public HObject evaluate(HFunction vect) {
			if (vect.size() == 2 && vect.get(1) instanceof HFunction) {
		 //			  && ((HFunction) vect.get(1)).isNumericNumber()) {
				HFunction	lst = (HFunction) vect.get(1);
				long			n;

				if (vect.get(0) instanceof HInteger) {
					HInteger num = (HInteger) vect.get(0);

					n = 1 << num.intValue();
				} else {
					if (vect.get(0) instanceof HDouble) {
						HDouble  dbl = (HDouble) vect.get(0);

						n = 1 << dbl.intValue();
					} else {
						return null;
					}
				}
				if (n < 2) {
					throw new HThrowException(C.Error, FFT,
													  new HString("number of samples must be >= 2."));
				}
				if (n > lst.size()) {
					throw new HThrowException(C.Error, FFT, 
													  new HString("calculated number of samples greater vector-length."));
				}
        HDoubleComplex arr[] = new HDoubleComplex[(int) n];

        HObject temp;
				for (int i = 0; i < (int) n; i++) {
          temp = C.NEV( lst.get(i) );
					if (temp instanceof HDouble) {
						arr[i] = new HDoubleComplex((HDouble) temp);
					} else {
						if (temp instanceof HDoubleComplex) {
							arr[i] = (HDoubleComplex) temp;
						} else {
							throw new HThrowException(C.Error, FFT,
															  new HString("only numerical values allowed."));
						}
					}
				}

				arr = FourierMath.transform(arr);

				HFunction	result = new HFunction(lst.head(), (int) n);

				for (int i = 0; i < (int) n; i++) {
					result.set(i, arr[i]);
				}

				return result;
			}

			return null;
		}

	}

	/**
	 * Class Declaration.
	 * 
	 * 
	 * @see
	 * 
	 * @author
	 * @version   %I%, %G%
	 */
	class EInverseFFT implements FunctionEvaluator {

		/**
		 * Method Declaration.
		 * 
		 * 
		 * @param vect
		 * 
		 * @return
		 * 
		 * @see
		 */
		public HObject evaluate(HFunction vect) {
			if (vect.size() == 2 && vect.get(1) instanceof HFunction) { 
	//				  && ((HFunction) vect.get(1)).isNumericNumber()) {
				long  n;

				if (vect.get(0) instanceof HInteger) {
					HInteger num = (HInteger) vect.get(0);

					n = 1 << num.intValue();
				} else {
					if (vect.get(0) instanceof HDouble) {
						HDouble  dbl = (HDouble) vect.get(0);

						n = 1 << dbl.intValue();
					} else {
						return null;
					}
				}

				HFunction	lst = (HFunction) vect.get(1);

				if (n < 2) {
					throw new HThrowException(C.Error, InverseFFT, 
													  new HString("length must be >= 2."));
				}
				if (n > lst.size()) {
					throw new HThrowException(C.Error, InverseFFT, 
													  new HString("calculated length greater vector-length."));
				}

				HDoubleComplex arr[] = new HDoubleComplex[(int) n];

        HObject temp;
				for (int i = 0; i < (int) n; i++) {
          temp = C.NEV( lst.get(i) );
					if (temp instanceof HDouble) {
						arr[i] = new HDoubleComplex((HDouble) temp);
					} else {
						if (temp instanceof HDoubleComplex) {
							arr[i] = (HDoubleComplex) temp;
						} else {
							throw new HThrowException(C.Error, InverseFFT, 
															  new HString("only numerical values allowed."));
						}
					}
				}

				arr = FourierMath.inverseTransform(arr);

				HFunction	result = new HFunction(lst.head(), (int) n);

				for (int i = 0; i < (int) n; i++) {
					result.set(i, arr[i]);
				}

				return result;
			}

			return null;
		}

	}

	/**
	 * Returns true if x is a power of 2.
	 * @author Don Cross
	 */

	/*
	 * private static boolean isPowerOf2(int x) {
	 * int	BITS_PER_WORD = 32;
	 * int	i, y;
	 * for (i = 1, y = 2; i < BITS_PER_WORD; i++, y <<= 1) {
	 * if (x == y) {
	 * return true;
	 * }
	 * }
	 * return false;
	 * }
	 */

	/**
	 * Constant: HSymbol "FFT" <a href="descript.htm#B-REF">(function-header)</a>
	 */
	final public HSymbol FFT = new HSymbol("FFT", new EFFT());

	/**
	 * Constant: HSymbol "InverseFFT" <a href="descript.htm#B-REF">(function-header)</a>
	 */
	final public HSymbol InverseFFT = new HSymbol("InverseFFT", 
			  new EInverseFFT());

	/**
	 * Class Constructor.
	 * 
	 * 
	 * @see
	 */
	public Fourier() {
		FFT.putSymbol();
		InverseFFT.putSymbol();
    C.resultOut.append("Loading Functions:\n"
								 + "FFT(<length>, {<double-complex>, ...})\n" 
								 + "InverseFFT(<length>, {<double-complex>, ...})\n");
			//{{INIT_CONTROLS
		//}}
}

	/**
	 * Method Declaration.
	 * 
	 * 
	 * @param vect
	 * 
	 * @return
	 * 
	 * @see
	 */
	public HObject evaluate(HFunction vect) {
		return C.Null;
	}

}



/*--- formatting done in "Sun Java Convention" style on 07-20-1999 ---*/
