/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package com.hartmath.expression;

import com.hartmath.util.*;
import com.hartmath.lib.*;

import java.lang.*;
import java.io.*;
import java.math.*;
import java.io.Serializable;

/**
 * HDouble represents a floating-point double value
 * @see HDoubleComplex
 * 
 * @author Klaus Hartlage  <A HREF="mailto:khartlage@t-online.de">khartlage@t-online.de</A>
 */
public class HDouble extends MathNumber implements HObject, HSignedNumber
{

   /**
    * The value of the Double.
    */
   private double               value;

   /**
    * Square root of 2.
    */
   public final static double   SQRT2 = 
      1.4142135623730950488016887242096980785696718753769;

   /**
    * Two times Pi.
    */
   public final static double   TWO_PI = 
      6.2831853071795864769252867665590057683943387987502;

   /**
    * Euler's gamma constant.
    */
   public final static double   EULERGAMMA = 
      0.57721566490153286060651209008240243104215933593992;

   /**
    * Catalan's constant.
    */
   public final static double   CATALAN =
      0.91596559417721901505460351493238411077414937428167;
   // Some IEEE machine constants

   /**
    * Relative machine precision.
    */
   public final static double   EPS = 2.22e-16;

   /**
    * The smallest positive floating-point number such that 1/xminin is machine representable.
    */
   public final static double   XMININ = 2.23e-308;

   /**
    * Returns true if the specified number is the special Not-a-Number (NaN) value.
    * @param v	the value to be tested
    */
   static public boolean isNaN(double v)
   {
      return Double.isNaN(v);
   }

   /**
    * Returns true if the specified number is infinitely large in magnitude.
    * @param v	the value to be tested
    */
   static public boolean isInfinite(double v)
   {
      return Double.isInfinite(v);
   }

   /**
    * Class Constructor.
    * 
    * 
    * @see
    */
   protected HDouble()
   {
      this.value = 0.0d;

      // {{INIT_CONTROLS
      // }}
   }

   /**
    * Constructs a Double wrapper for the specified double value.
    * @param value the initial value of the double
    */
   public HDouble(double value)
   {
      this.value = value;
   }

   /**
    * Constructs a Double object initialized to the value specified by the
    * String parameter.
    * @param s		the String to be converted to a Double
    * @exception	NumberFormatException If the String does not contain a parsable number.
    */
   public HDouble(String s) throws NumberFormatException
   {

      // REMIND: this is inefficient

      this(Double.valueOf(s).doubleValue());
   }

   /**
    * Returns true if this Double value is the special Not-a-Number (NaN) value.
    */
   public boolean isNaN()
   {
      return isNaN(value);
   }

   /**
    * Returns true if this Double value is infinitely large in magnitude.
    */
   public boolean isInfinite()
   {
      return Double.isInfinite(value);
   }

   /**
    * Returns a String representation of this Double object.
    */
   public String toString()
   {
      return Double.toString(value);
   }
   
   public void toStringBuffer(StringBuffer buf)
   {
     buf.append( Double.toString(value) );
   }
   /**
    * Returns the integer value of this Double (by casting to an int).
    */
   public int intValue()
   {
      return (int) value;
   }

   /**
    * Returns the long value of this Double (by casting to a long).
    */
   public long longValue()
   {
      return (long) value;
   }

   /**
    * Returns the float value of this Double.
    */
   public float floatValue()
   {
      return (float) value;
   }

   /**
    * Returns the double value of this Double.
    */
   public double doubleValue()
   {
      return (double) value;
   }

   /**
    * Returns a hashcode for this Double.
    */
   public int hashCode()
   {
      return (int) value;
   }

   /**
    * Compares this object against the specified object.
    * <p>
    * <em>Note: To be useful in hashtables this method
    * considers two NaN double values to be equal. This
    * is not according to IEEE specification</em>
    * 
    * @param obj		the object to compare with
    * @return 		true if the objects are the same; false otherwise.
    */
   public boolean equals(Object obj)
   {
      return ((obj instanceof HDouble) 
              && (Double.doubleToLongBits(((HDouble) obj).value) 
                  == Double.doubleToLongBits(value)));
   }

   /**
    * Returns the double-float corresponding to a given bit represention.
    */

   // public static native double longBitsToDouble(long bits);

   public boolean isEqual(HDouble b)
   {
      return (Double.doubleToLongBits(b.value) 
              == Double.doubleToLongBits(value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param b
    * 
    * @return
    * 
    * @see
    */
   public boolean isUnequal(HDouble b)
   {
      return !isEqual(b);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param b
    * 
    * @return
    * 
    * @see
    */
   public boolean isLess(HDouble b)
   {
      return value < b.value;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param b
    * 
    * @return
    * 
    * @see
    */
   public boolean isLessEqual(HDouble b)
   {
      return value <= b.value;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param b
    * 
    * @return
    * 
    * @see
    */
   public boolean isGreater(HDouble b)
   {
      return value > b.value;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param b
    * 
    * @return
    * 
    * @see
    */
   public boolean isGreaterEqual(HDouble b)
   {
      return value >= b.value;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public boolean isZero()
   {
      return (value == 0.0d);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public boolean isNonZero()
   {
      return (value != 0.0d);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public boolean matches(HObject obj)
   {
      return equals(obj);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public boolean unequals(Object obj)
   {
      return !equals(obj);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public boolean less(Object obj)
   {
      if (obj instanceof HDouble)
      {
         return (value < ((HDouble) obj).value);
      }

      if (obj instanceof HObject)
      {
         return (hierarchy() < ((HObject) obj).hierarchy());
      }

      return false;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public boolean greater(Object obj)
   {
      if (obj instanceof HObject)
      {
         return ((HObject) obj).less(this);
      }

      return false;
   }

   /**
    * Returns true if <code>this</code> can be evaluated.
    * <code>obj</code> is the evaluation result.
    * @param obj         the (new) evaluated object
    */
   public HObject evaluate()
   {
      //Context   ct = ((ContextThread) Thread.currentThread()).getContext();

      if (!C.numericFlag)
      {
         C.numericFlag = true;

         return this;
      }

      return null;
   }

   /**
    * Returns true if <code>this</code> is (pattern-) matches the object
    * <code>obj</code>
    * @param obj         an object to compare with
    * @param subst       the substitution for the object
    */
   public HObject substitute(HObject obj, HObject subst)
   {
      if (equals(obj))
      {
         return subst;
      }

      return this;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public HObject substitutePattern()
   {
      return this;
   }

   /**
    * Returns an specific integer value for each class in the math-objects hierarchy
    */
   final public int hierarchy()
   {
      return HObject.DOUBLEID;          // 2
   }

   /**
    * Write the double value to an DataOutputStream
    * 
    * @param os
    * 
    * @exception IOException
    */   /*
   public void writeTo(DataOutputStream os) throws IOException
   {
      os.writeByte(HObject.DOUBLEFILEID);
      os.writeDouble(value);
   }       */

   /**
    * Return C.Double as the header for HDouble-objects.
    * 
    * @return C.Double
    * 
    * @see C
    */
   public HSymbol head()
   {
      return C.Double;
   }

   /**
    * Method declaration
    *
    *
    * @param head
    *
    * @return
    *
    * @see
    */
   public HObject apply(HObject head)
   {
      if (C.DEBUG)
      {         // only in debug mode
         if (head == null)
         {
            throw new HThrowException(C.HeadFormatError, C.Symbol, C.Null);
         }
      }

      if (head instanceof HSymbol)
      {
         return C.EV(new HFunction((HSymbol) head, this));
      }

      if ((head instanceof HFunction) && ((HFunction) head).isPureFunction())
      {
         HFunction      l = (HFunction) ((HFunction) head).clone();

         l.add(this);

         return C.EV(l);
      }

      throw new HThrowException(C.HeadFormatError, C.Lambda, head);
   }

   /**
    * Return <code>true</code> if <code>this</code> is a List
    * 
    * @return <code>false</code>
    * 
    * @see HFunction
    */
   public boolean isList()
   {
      return false;
   }

   /**
    * Return <code>true</code> if <code>this</code> is a number
    * 
    * @return <code>true</code>
    * 
    * @see HFunction
    */
   public boolean isNumber()
   {
      return true;
   }

   /**
    * Return <code>true</code> if <code>this</code> ia a rational number
    * 
    * @return <code>false</code>
    * 
    * @see HFunction
    */
   public boolean isRational()
   {
      return false;
   }

   /**
    * Return <code>true</code> if the argument is a subobject or equal to <code>this</code>
    * 
    * @return <code>this.equals(obj)</code>
    * 
    * @see HFunction
    */
   public boolean isMember(HObject obj)
   {
      return equals(obj);
   }

   /**
    * Return <code>true</code> if the argument equals the head of <code>this</code> or any subobject head
    * 
    * @see HFunction
    */
   public boolean isHeadMember(HSymbol sym)
   {
      return C.Double.equals(sym);
   }

   /**
    * Return <code>true</code> if <code>this</code> > 0.0
    * 
    * @return <code>this > 0.0</code>
    * 
    * @see HFunction
    */
   public boolean isPositive()
   {
      return (value > 0.0d);
   }

   /**
    * Return <code>true</code> if <code>this</code> < 0.0
    * 
    * @return <code>this < 0.0</code>
    * 
    * @see HFunction
    */
   public boolean isNegative()
   {
      return (value < 0.0d);
   }

   /**
    * Return <code>true</code> if <code>this</code> >= 0.0
    * 
    * @return <code>this >= 0.0</code>
    * 
    * @see HFunction
    */
   public boolean isNonNegative()
   {
      return (value >= 0.0d);
   }

   /**
    * Returns the absolute value.
    * If the argument is not negative, the argument is returned.
    * If the argument is negative, the negation of the argument is returned.
    * 
    * @return  the absolute value of the object.
    * @see java.lang.Math
    */
   public HObject abs()
   {
      return new HDouble(Math.abs(value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @exception java.lang.ArithmeticException
    * 
    * @see
    */
   public HDouble add(HDouble parm1) throws java.lang.ArithmeticException
   {
      return new HDouble(value + parm1.value);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @exception java.lang.ArithmeticException
    * 
    * @see
    */
   public HDouble divide(HDouble parm1) throws java.lang.ArithmeticException
   {
      return new HDouble(value / parm1.value);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @see
    */
   public HDouble max(HDouble parm1)
   {
      return new HDouble(Math.max(value, parm1.value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @see
    */
   public HDouble min(HDouble parm1)
   {
      return new HDouble(Math.min(value, parm1.value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @see
    */
   public HDouble mod(HDouble parm1)
   {
      return new HDouble(value % parm1.value);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @see
    */
   public HDouble IEEERemainder(HDouble parm1)
   {
      return new HDouble(Math.IEEEremainder(value, parm1.value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @see
    */
   public HDouble multiply(HDouble parm1)
   {
      return new HDouble(value * parm1.value);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public HDouble negate()
   {
      return new HDouble(-1.0 * value);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @exception java.lang.ArithmeticException
    * 
    * @see
    */
   public HDouble pow(HDouble parm1) throws java.lang.ArithmeticException
   {
      return new HDouble(Math.pow(value, parm1.value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param d0
    * 
    * @return
    * 
    * @see
    */
   public static HDouble exp(double d0)
   {
      return new HDouble(Math.exp(d0));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public HDouble exp()
   {
      return new HDouble(Math.exp(value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param d0
    * 
    * @return
    * 
    * @see
    */
   public static HDouble sin(double d0)
   {
      return new HDouble(Math.sin(d0));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public HDouble sin()
   {
      return new HDouble(Math.sin(value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param d0
    * 
    * @return
    * 
    * @see
    */
   public static HDouble cos(double d0)
   {
      return new HDouble(Math.cos(d0));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public HDouble cos()
   {
      return new HDouble(Math.cos(value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param d0
    * 
    * @return
    * 
    * @see
    */
   public static HDouble tan(double d0)
   {
      return new HDouble(Math.tan(d0));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public HDouble tan()
   {
      return new HDouble(Math.tan(value));
   }

   /**
    * Returns the hyperbolic sine of a double.
    * @param d0 a double
    */
   public static HDouble sinh(double d0)
   {
      return new HDouble(ExtraMath.sinh(d0));
   }

   /**
    * Returns the hyperbolic sine of this HDouble.
    */
   public HDouble sinh()
   {
      return new HDouble(ExtraMath.sinh(value));
   }

   /**
    * Returns the hyperbolic cosine of a double.
    * @param d0 a double
    */
   public static HDouble cosh(double d0)
   {

      // return new HDouble((Math.exp(d0) + Math.exp(-d0)) / 2.0);
      return new HDouble(ExtraMath.cosh(d0));
   }

   /**
    * Returns the hyperbolic cosine this HDouble.
    */
   public HDouble cosh()
   {

      // return new HDouble((Math.exp(value) + Math.exp(-value)) / 2.0);
      return new HDouble(ExtraMath.cosh(value));
   }

   /**
    * Returns the hyperbolic tangent of this HDouble
    */
   public HDouble tanh()
   {

      // return sinh().divide(cosh());
      return new HDouble(ExtraMath.tanh(value));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param parm1
    * 
    * @return
    * 
    * @see
    */
   public HDouble subtract(HDouble parm1)
   {
      return new HDouble(value - parm1.value);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public int precedence()
   {
      if (value < 0)
      {
         return 100;
      }

      return 999999;
   }

   /**
    * 0 for this == 0
    * +1 for re(this) > 0 or ( re(this)==0 and im(this)>0 )
    * -1 for re(this) < 0 or ( re(this)==0 and im(this)<0 )
    */
   public int compSignum()
   {
      return signum();
   }

   /**
    * Returns the signum function of this number (i.e., -1, 0 or 1 as
    * the value of this number is negative, zero or positive).
    */

   // public int signum() {
   // return numer.signum();
   // }

   public int signum()
   {
      if (value > 0)
      {
         return 1;
      }

      if (value == 0)
      {
         return 0;
      }

      return -1;
   }

   /**
    * Returns the smallest (closest to negative infinity)
    * <code>HObject</code> value that is not less than <code>this</code> and is
    * equal to a mathematical integer.
    * 
    * @return  the smallest (closest to negative infinity)
    * <code>HObject</code> value that is not less than <code>this</code>
    * and is equal to a mathematical integer.
    */
   public HObject ceil()
   {
      return new HDouble(Math.ceil(value));
   }

   /**
    * Returns the largest (closest to positive infinity)
    * <code>HObject</code> value that is not greater than <code>this</code> and
    * is equal to a mathematical integer.
    * 
    * @return  the largest (closest to positive infinity)
    * <code>HObject</code> value that is not greater than <code>this</code>
    * and is equal to a mathematical integer.
    */
   public HObject floor()
   {
      return new HDouble(Math.floor(value));
   }

   /**
    * returns the conjugate complex of a number
    */
   public HObject conjugate()
   {
      return this;
   }

   /**
    * returns the real part of a number
    */
   public HObject re()
   {
      return this;
   }

   /**
    * returns the imaginary part of a number
    */
   public HObject im()
   {
      return C.CD0;
   }

   /**
    * Returns the arc hyperbolic sine of a this HDouble.
    */
   public HDouble asinh()
   {
      return new HDouble(ExtraMath.asinh(value));
   }

   /**
    * Returns the arc hyperbolic cosine of this HDouble.
    */
   public HDouble acosh()
   {
      return new HDouble(ExtraMath.acosh(value));
   }

   /**
    * Returns the arc hyperbolic tangent of a HDouble.
    * @param value a HDouble
    */
   public HDouble atanh()
   {
      return new HDouble(ExtraMath.atanh(value));
   }

   /**
    * Returns the addition of this number and another.
    */
   public MathNumber add(final MathNumber n)
   {
      if (n instanceof HDouble)
         return add((HDouble) n);

         // else if(n instanceof MathInteger)
         // return add(new HDouble(((MathInteger)n).value()));
      else
         throw new IllegalArgumentException("Number class not recognised by this method.");
   }

   /**
    * Returns the subtraction of this number and another.
    */
   public MathNumber subtract(final MathNumber n)
   {
      if (n instanceof HDouble)
         return subtract((HDouble) n);

         // else if(n instanceof MathInteger)
         // return subtract(new HDouble(((MathInteger)n).value()));
      else
         throw new IllegalArgumentException("Number class not recognised by this method.");
   }

   /**
    * Returns the multiplication of this number and another.
    */
   public MathNumber multiply(final MathNumber n)
   {
      if (n instanceof HDouble)
         return multiply((HDouble) n);

         // else if(n instanceof MathInteger)
         // return multiply(new HDouble(((MathInteger)n).value()));
      else
         throw new IllegalArgumentException("Number class not recognised by this method.");
   }

   /**
    * Method declaration
    *
    *
    * @param n
    *
    * @return
    *
    * @see
    */
   public MathNumber divide(final MathNumber n)
   {
      if (n instanceof HDouble)
         return divide((HDouble) n);

         // else if(n instanceof MathInteger)
         // return divide(new HDouble(((MathInteger)n).value()));
      else
         throw new IllegalArgumentException("Number class not recognised by this method.");
   }

   /**
    * Returns the double value.
    */
   public double value()
   {
      return value;
   }

}



