/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package com.hartmath.expression;

import com.hartmath.lib.*;
import com.hartmath.util.*;
import com.hartmath.patternmatching.*;
import com.hartmath.mapping.*;

//import org.kaffe.java.util.*;
import org.kaffe.java.lang.UnsupportedOperationException;

import java.util.Enumeration;
import java.lang.*;
import java.io.*;


/**
 * A function represents a mathematical expression.
 * It is constructed from a symbolic header which is representing
 * the function name, and arbitrary number of HObject's as arguments.
 * Ex.: the mathematical expression <code>Sin( Pi^2 )</code> is constructed with the following statement:
 * new HFunction(C.Sin, new HFunction( C.Pow, C.Pi, C.C2) )
 * Alternatively you can construct functions for "predefined" symbols with the HSymbol.f() method in the following way:
 * C.Sin.f( C.Pow.f( C.Pi, C.C2) )
 * 
 * @see HSymbol
 * @see HObject
 * @see C
 * 
 * @author Klaus Hartlage  <A HREF="mailto:khartlage@t-online.de">khartlage@t-online.de</A>
 */
public class HFunction extends HArrayList implements HObject, Cloneable
{
   final static public int      NOPROPERTY = 0xFF00,    // default property
   MATCHEQUAL = 
      0x0001,           // match the left-hand-side of a rule with method equals()
   MATCHSEQUENCE = 
      0x0002,           // match the left-hand-side of a rule with method matches()
   MATRIXFORM = 0x0100, FLAT_EVALED = 0x0010, ORDERLESS_EVALED = 0x0020;

   /**
    * No attribute; this flag indicates that a rule should be written
    * as <code>lhs = rhs</code>
    * 
    * @see EDefinition
    */
   final static public int      SET = 0x1000;

   /**
    * No attribute; this flag indicates that a rule should be written
    * as <code>lhs := rhs</code>
    * 
    * @see EDefinition
    */
   final static public int      SETDELAYED = 0x2000;

   /*
    * Properties of this function
    */
   int                          prop;

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   final public int getProperties()
   {
      return prop;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param att
    * 
    * @see
    */
   final public void setProperties(int att)
   {
      prop |= att;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param att
    * 
    * @see
    */
   final public void clearProperties(int att)
   {
      prop &= (att & 0xFFFF);
   }

   /*
    * The header symbol of a function
    */
   HSymbol      symbol;

   /**
    * Constructs a function with the Elements of the
    * function <code>f</code> as arguments;
    * 
    * @param <code>sym</code> is the new head of the function (i.e. s=f)
    * @param <code>first</code> and <code>last</code> is the range of arguments;
    * @param <code>incl</code> is a flag for "range include" (true) or
    * "range exclude" (false)
    */
   public HFunction(HFunction f, HSymbol sym, int first, int last, 
                    boolean incl)
   {
      symbol = sym;

      if (incl == true)
      {         // range include
         for (int i = first; i < last; i++)
         {
            add(f.a[i]);
         }
      }
      else
      {         // range exclude
         for (int i = 0; i < first; i++)
         {
            add(f.a[i]);
         }

         for (int j = last; j < f.len; j++)
         {
            add(f.a[j]);
         }
      }
   }

   /**
    * Constructs an empty function with the specified storage capacity.
    * 
    * @param s is the head of the function (i.e. s=sin ==>  Sin(...))
    * @param initialCapacity the initial storage capacity of the vector
    */
   public HFunction(HSymbol s, int initialCapacity)
   {
      super(initialCapacity, true);

      symbol = s;
   }

   /**
    * Constructor declaration
    * 
    * 
    * @param rowSym
    * @param colSym
    * @param initialRowCapacity
    * @param initialColCapacity
    * @param p
    * 
    * @see
    */
   public HFunction(HSymbol rowSym, HSymbol colSym, int initialRowCapacity, 
                    int initialColCapacity, int p)
   {
      super(initialRowCapacity, true);
      
      for (int i = 0; i < initialRowCapacity; i++)
      {
         super.a[i] = new HFunction(colSym, initialColCapacity);
      }
      
      prop = p;
      symbol = rowSym;
   }

   /**
    * Constructs a function with no arguments
    * 
    * @param s is the head of the function (i.e. s=sin ==>  Sin(...))
    */
   public HFunction(HSymbol s)
   {
      symbol = s;

   }

   /**
    * Constructs a HFunction with 1 argument
    * 
    * @param s is the head of the function (i.e. s=sin ==>  Sin(...))
    * @param obj1 the first argument of the function  (i.e. Sin(<obj1>)
    */
   public HFunction(HSymbol s, HObject obj1)
   {
      super(1,true);

      super.a[0]=obj1;

      symbol = s;
   }

   /**
    * Constructs a function with 2 arguments
    * 
    * @param s is the head of the function (i.e. s=f)
    * @param obj1, obj2 are the arguments of the function
    * (i.e. f(<obj1>, <obj2>))
    */
   public HFunction(HSymbol s, HObject obj1, HObject obj2)
   {
      super(2,true);

      super.a[0]=obj1;
      super.a[1]=obj2;

      symbol = s;
   }

   /**
    * Constructs a function with 3 arguments
    * 
    * @param s is the head of the function (i.e. s=f)
    * @param obj1, obj2, obj3 are the arguments of the function
    * (i.e. f(<obj1>, <obj2>,<obj3>))
    */
   public HFunction(HSymbol s, HObject obj1, HObject obj2, HObject obj3)
   {
      super(3,true);
      
      super.a[0]=obj1;
      super.a[1]=obj2;
      super.a[2]=obj3;

      symbol = s;
   }

   /**
    * Constructs a function with 4 arguments
    * 
    * @param s is the head of the function (i.e. s=f)
    * @param obj1, obj2, obj3, obj4 are the arguments of the function
    * (i.e. f(<obj1>, <obj2>,<obj3>,<obj4>))
    */
   public HFunction(HSymbol s, HObject obj1, HObject obj2, HObject obj3, 
                    HObject obj4)
   {
      super(4, true);

      super.a[0]=obj1;
      super.a[1]=obj2;
      super.a[2]=obj3;            
      super.a[3]=obj4;

      symbol = s;
   }

   /**
    * Constructs a function with the Elements of the array as arguments
    * 
    * @param s is the head of the function (i.e. s=f)
    * @param v contains the arguments of the function
    */
   public HFunction(HSymbol s, HArrayList v)
   {
      super(v);

      symbol = s;
   }
   //public HFunction(HSymbol s, Array arr)
   //{
   //   symbol = s;
   //
   //   for (int i = 0; i < arr.len; i++)
   //   {
   //      add(arr.a[i]);
   //   }
   //}
   public HFunction(HSymbol s, double[] arr)
   {
      super(arr.length, true);

      for (int i = 0; i < arr.length; i++)
      {
         a[i] = new HDouble(arr[i]);
      }
      symbol = s;
   }

   /**
    * Constructor declaration
    * 
    * 
    * @param s
    * @param arr
    * 
    * @see
    */
   public HFunction(HSymbol s, HDoubleComplex[] arr)
   {
      super(arr.length, true);
      for (int i = 0; i < arr.length; i++)
      {
         a[i] = arr[i];
      }
      symbol = s;
   }
   /**
    * Make a shallow copy of this function.
    * 
    * Only the structure on the top-level of the function will be copied;
    * The atomic objects of the HFunction will not be "cloned".
    */
   final public Object clone()
   {
   	 HFunction fun = (HFunction) super.clone();
   	 fun.symbol = symbol;
   	 return fun;
   }

   /**
    * Make a deep copy of this function until level.
    * 
    * The structure of all levels of the function will be copied;
    * The atomic objects of the HFunction will not be "deepCopied".
    * 
    * @param level is the level of argument which should be copied
    * Use Integer.MAX_VALUE if you want to copy all levels
    */
   public HFunction deepCopy(int level)
   {
      HFunction f = null;

      f = (HFunction) clone();
      level--;

      if (level > 0)
      {
         for (int i = 0; i < len; i++)
         {
            if ((get(0) instanceof HFunction))
            {
               f.set(i, ((HFunction) get(i)).deepCopy(level));
            }
            else
            {
               f.set(i, get(i));
            }
         }
      }

      return f;
   }

   /**
    * Get a list of the arguments as a StringBuffer
    * 
    * @param buf
    * @param start
    * @param fullForm
    * 
    * @see
    */
   void argumentsToString(StringBuffer buf, int start, boolean fullForm)
   {
      for (int i = start; i < len; i++)
      {
         if (C.DEBUG)
         {
            if (a[i] == null)
            {
               buf.append("<null>");

               continue;
            }
         }

         if (fullForm && a[i] instanceof HFunction)
         {
            buf.append(((HFunction) a[i]).toFullForm());
         }
         else
         {
            a[i].toStringBuffer(buf);
         }

         if (i != (len - 1))
         {
            buf.append(',');

            if ((prop & MATRIXFORM) == MATRIXFORM)
            {
               buf.append("\n ");
            }
         }
      }
   }

   /**
    * Convert the function to a String-Object without operator notation
    * 
    * @return  the function as a string [i.e: "f(x1, x2,...)" or "Add(a,b,c)", "List(a,b,c)"]
    */
   public String toFullForm()
   {
      StringBuffer      buf = new StringBuffer();

      if (C.DEBUG)
      {
         if (symbol == null)
         {
            buf.append("<symbol-null>");
         }
      }

      buf.append(symbol.toString());
      buf.append('(');
      argumentsToString(buf, 0, true);
      buf.append(')');

      return buf.toString();
   }

   /**
    * Convert the function to a String-Object
    * 
    * @return  the function as a string [i.e: "f(x1, x2,...)" or "a+b+c"]
    */
   public String toString()
   {
      StringBuffer      buf = new StringBuffer();
      toStringBuffer(buf);
      return buf.toString();
   }

   public void toStringBuffer(StringBuffer buf)
   {
      // check if the function is written in operator notation
      try
      {

         if (checkAttributes(HSymbol.OPERATOR))
         {
            buf.append(symbol.toOpString(this));
            return;
         }

         if (C.DEBUG)
         {   /*
            if (C.traceFlag)
            {
               if (prop != NOPROPERTY)
               {
                  buf.append('(');
                  buf.append(new Integer(prop).toString());
                  buf.append(')');
               }
            }  */
         }

         if (symbol == C.Slot)
         {
            if (len == 1 && a[0] instanceof HInteger)
            {
               buf.append('#');

               if (!a[0].equals(C.C1))
               {
                  a[0].toStringBuffer(buf);
               }
               return;
            }
         }
         if (symbol == C.HoldForm)
         {
            if (len == 1)
            {
               a[0].toStringBuffer(buf);
               return;
            }
         }
         if (symbol == C.Lambda)
         {
            if (len >= 2 && a[0].equals(C.Null))
            {
               buf.append('(');
               a[1].toStringBuffer(buf);
               buf.append(")&");

               if (len > 2)
               {
                  buf.append('(');
                  argumentsToString(buf, 2, false);
                  buf.append(')');
               }

               return;
            }
         }

         if (symbol == C.Part)
         {

            // special character for head C.Part

            if (len != 0)
            {
               if ((a[0] instanceof HFunction) 
                   && ((HFunction) a[0]).checkAttributes(HSymbol.OPERATOR))
               {
                  buf.append('(');
                  a[0].toStringBuffer(buf);
                  buf.append(')');
               }
               else
               {
                  a[0].toStringBuffer(buf);
               }
            }

            buf.append('[');
            argumentsToString(buf, 1, false);
            buf.append(']');

            return;
         }

         if (symbol == C.List)
         {

            // special character for head C.List

            buf.append('{');
         }
         else
         {
            symbol.toStringBuffer(buf);
            buf.append('(');
         }

         argumentsToString(buf, 0, false);

         if (symbol == C.List)
         {
            buf.append('}');
         }
         else
         {
            buf.append(')');
         }
      }
      catch (Throwable se)
      {
         C.resultOut.appendLine("Exception occured: " + se.getMessage());

         if (C.DEBUG)
         {
            se.printStackTrace();
         }
      }
   }

   
   /**
    * the intValue of the object which is stored at the given index
    * the object has to be an instance of HInteger or HDouble;
    * otherwise the function returns Integer.MIN_VALUE
    */
   public int toInt(int index)
   {
      int       n = Integer.MIN_VALUE;

      if (a[index] instanceof HInteger)
      {
         n = ((HInteger) a[index]).intValue();
      }
      else
      {
         if (a[index] instanceof HDouble)
         {
            n = ((HDouble) a[index]).intValue();
         }
      }

      return n;
   }

   /**
    * Get the attributes of a function
    * 
    * The attributes are inherited from HFunction.symbol
    */
   final public int getAttributes()
   {
      return symbol.getAttributes();
   }

   /**
    * Check the attributes of a function
    * 
    * The attributes are inherited from HFunction.symbol
    */
   final public boolean checkAttributes(int att)
   {
      return symbol.checkAttributes(att);
   }

   /**
    * Determine and init all pattern symbols in this function
    */
   public HArrayList initPattern(HArrayList result)
   {
      for (int i = 0; i < len; i++)
      {
         if (a[i] instanceof HFunction)
         {
            result = ((HFunction) a[i]).initPattern(result);
         }

         if (a[i] instanceof HPattern)
         {
            if (((HPattern) a[i]).isEmpty() 
                || ((HPattern) a[i]).getValue() != null)
            {
               ((HPattern) a[i]).initValue();
               result.add(a[i]);
            }
         }
      }

      return result;
   }

   /**
    * Reset all pattern symbols of the HArrayList
    */
   public static void resetPattern(HArrayList result)
   {
      for (int i = 0; i < result.len; i++)
      {
         ((HPattern) result.a[i]).resetValue();
      }
   }

   /**
    * Get all pattern symbols in this function
    */
   public void getPattern(HArrayList pats)
   {
      boolean   flag;

      for (int i = 0; i < len; i++)
      {
         if (a[i] instanceof HFunction)
         {
            ((HFunction) a[i]).getPattern(pats);

            continue;
         }

         if (a[i] instanceof HPattern)
         {
            if (((HPattern) a[i]).getValue() == null)
            {
               flag = true;

               for (int j = 0; j < pats.len; j++)
               {
                  if (a[i].equals(pats.a[j]))
                  {
                     flag = false;

                     break;
                  }
               }

               if (flag)
               {
                  pats.add(a[i]);
               }
            }
         }
      }
   }

   /**
    * Clear all pattern symbols of the given HArrayList
    */
   public static void clearPattern(HArrayList result)
   {
      for (int i = 0; i < result.len; i++)
      {
         ((HPattern) result.a[i]).setValue(null);
      }
   }

   /**
    * Check if this function represents a vector
    * 
    * @return  null if this is not a vector
               {<vector-length>, <vector-type>} otherwise
               <vector-type> is one of:
               -1: HDoubleComplex vector
                1: HDouble vector
                0: all other vectors
    */
   public int[] isVector()
   {
      int ntype = 0;
      if (head() == C.List)
      {
         //Context ct = ((ContextThread) Thread.currentThread()).getContext();
         if (C.numericFlag) {
           ntype = 1;
         for (int i = 0; i < len; i++)
         {
            if ((a[i] instanceof HFunction)
                && ( (a[i]).head() == C.List))
            {
               return null;
            }
            if (a[i] instanceof HDoubleComplex) {
              if (ntype > 0) {
                ntype = -1;
              }
            }
            else {
              if (! (a[i] instanceof HDouble) ) {
                ntype = 0;
              }
            }
         }
         }
         else {
         for (int i = 0; i < len; i++)
         {
            if ((a[i] instanceof HFunction)
                && ((a[i]).head() == C.List))
            {
               return null;
            }
         }
         }

         return new int[]
               {
                  len, ntype
               };
      }

      return null;
   }

   /**
    * Check if this function represents a matrix
    * 
    * @return  array of row-length and column-length and numericType 0 -> not numeric 1-> double -1 -> complex
    */
   public int[] isMatrix()
   {
      int       cols = 0;
      int       rows = len;

      // numericType:
      int       ntype = 0;

      if ((head() == C.List) && (rows > 0))
      {
         if (a[0] instanceof HFunction)
         {
            HFunction   g = (HFunction) a[0];
            //Context ct = ((ContextThread) Thread.currentThread()).getContext();
            if (g.head() == C.List)
            {
               cols = g.len;

               if (C.numericFlag)
               {
                  ntype = 1;

                  HFunction     rw;
                  //attention index >= 0
                  for (int i = 0; i < rows; i++)
                  {
                     if ((!(a[i] instanceof HFunction))
                         || ((a[i]).head() != C.List) 
                         || (cols != ((HFunction) a[i]).len))
                     {
                        return null;
                     }

                     rw = (HFunction) a[i];

                     for (int j = 0; j < cols; j++)
                     {
                        if (rw.a[j] instanceof HDoubleComplex)
                        {
                           if (ntype > 0)
                           {
                              ntype = -1;
                           }
                        }
                        else
                        {
                           if (!(rw.a[j] instanceof HDouble))
                           {

                              // not a numeric matrix
                              ntype = 0;
                           }
                        }
                     }
                  }
               }
               else
               {
               //attention index >= 1
                  for (int i = 1; i < rows; i++)
                  {
                     if ((!(a[i] instanceof HFunction))
                         || ((a[i]).head() != C.List) 
                         || (cols != ((HFunction) a[i]).len))
                     {
                        return null;
                     }
                  }
               }

               return new int[]
               {
                  rows, cols, ntype
               };
            }
         }
      }

      return null;
   }
   /**
    * @see     java.lang.Object.hashCode()
    */
   public int hashCode()
   {
      if (len == 0)
      {
         return (symbol.hashCode() | 0x7AAAAAAA);
      }

      return ((symbol.hashCode()<<16)  | ( a[0]).head().hashCode());
   }

   /**
    * @return a special hashcode for pattern-matching
    */
   public int patternHashCode()
   {
      return ((symbol.hashCode()<<16) | C.Pattern.hashCode());
   }

   /**
    * @see     java.lang.Object.equals()
    */
   public boolean equals(Object obj)
   {
      if (obj instanceof HFunction)
      {
         HFunction      temp = (HFunction) obj;

         if (len != temp.len)
         {
            return false;
         }

         if (symbol.compareTo(temp.symbol) != 0)
         {
            return false;
         }

         for (int i = 0; i < len; i++)
         {
            if (!a[i].equals(temp.a[i]))
            {
               return false;
            }
         }

         return true;
      }

      return false;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public boolean unequals(Object obj)
   {
      return !equals(obj);
   }

   /**
    * this < obj ?
    */
   public boolean less(Object obj)
   {
      if (obj instanceof HFunction)
      {
         HFunction      temp = (HFunction) obj;
         int            sz = len;

         if (sz > temp.len)
         {
            sz = temp.len;
         }

         int    cp = symbol.compareTo(temp.symbol);

         if (cp < 0)
         {
            return true;
         }

         if (cp > 0)
         {
            return false;
         }

         if ((symbol.getAttributes() & HSymbol.ORDERLESS) 
             == HSymbol.ORDERLESS)
         {
            int j = len - 1;
            int k = temp.len - 1;

            for (int i = 0; i < sz; i++)
            {
               if (( a[j]).less( temp.a[k]))
               {
                  return true;
               }

               if (( temp.a[k]).less( a[j]))
               {
                  return false;
               }

               j--;
               k--;
            }
         }
         else
         {
            for (int i = 0; i < sz; i++)
            {
               if (( a[i]).less( temp.a[i]))
               {
                  return true;
               }

               if (( temp.a[i]).less( a[i]))
               {
                  return false;
               }
            }
         }

         return len < temp.len;
      }

      if (obj instanceof HObject)
      {
         return (hierarchy() < ((HObject) obj).hierarchy());
      }

      return false;
   }

   /**
    * this > obj ?
    */
   public boolean greater(Object obj)
   {
      if (obj instanceof HObject)
      {
         return ((HObject) obj).less(this);
      }

      return false;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param from0
    * @param f1
    * @param from1
    * 
    * @return
    * 
    * @see
    */
   public boolean equalsFromPosition(int from0, HFunction f1, int from1)
   {
      if ((len - from0) != (f1.len - from1))
      {
         return false;
      }

      int       j = from1;

      for (int i = from0; i < len; i++)
      {
         if (!a[i].equals(f1.a[j++]))
         {
            return false;
         }
      }

      return true;
   }

   /**
    * Thread through all lists in the arguments of the function
    * [i.e. the function header has the attribute HSymbol.LISTABLE]
    * 
    * example: Sin({2,x,Pi}) ==> {Sin(2),Sin(x),Sin(Pi)}
    * @see HSymbol
    */
   private HFunction threadList(int listLength)
   {
      if (C.DEBUG) {
        if (listLength == 0)
        {
          throw new UnsupportedOperationException();
        }
      }

      HFunction res0 = new HFunction(C.List, listLength);

      for (int j = 0; j < listLength; j++)
      {
         HFunction      res1 = new HFunction(symbol, len);

         for (int i = 0; i < len; i++)
         {
            if (( a[i]).isList())
            {
               HFunction arg = (HFunction) a[i];

               res1.a[i] = arg.a[j];
            }
            else
            {
               res1.a[i] = a[i];
            }
         }      // end for j

         res0.a[j] = res1;
      }         // end for i

      return res0;
   }

   /**
    * Flatten the function
    * [i.e. the function header has the attribute HSymbol.FLAT]
    * 
    * example: suggest the Symbol f has the attribute HSymbol.FLAT
    * f(a,b,f(x,y,f(u,v)),z) ==> f(a,b,x,y,u,v,z)
    * @see HSymbol
    */
   public HFunction flatten()
   {
      HFunction res = new HFunction(symbol);
      boolean   isEvaled = false;

      for (int i = 0; i < len; i++)
      {
         if ((a[i] instanceof HFunction) 
             && (symbol.equals(( a[i]).head())))
         {
         	  isEvaled = true;
            res.addAll( (HFunction) a[i] );
            /*
            HFunction   arg = ((HFunction) a[i]);
            res.addAll( (HFunction) a[i]);
            for (int j = 0; j < arg.len; j++)
            {
               res.add(arg.a[j]);
            }           // end for
             */
            isEvaled = true;
         }
         else
         {
            res.add( a[i]);
         }
      }

      if (isEvaled)
      {
         return res;
      }

      res = null;

      return null;
   }

   /**
    * Sort the function
    * [i.e. the function header has the attribute HSymbol.ORDERLESS]
    * 
    * example: suggest the Symbol f has the attribute HSymbol.ORDERLESS
    * f(a,b,x,y,u,v,z) ==> f(a,b,u,v,x,y,z)
    * @see HSymbol
    */

   // a simple shakerSort:

   /*
    * Object s1, s2;
    * boolean evaled = false;
    * int  j, k, l, r;
    * if (len > 1) {
    * l = 1;
    * r = len - 1;
    * k = r;
    * do {
    * for(j=r;j>=l;j--) {
    * s1 = at(j-1);
    * s2 = a[j];
    * if ( ((HObject)s2).less( (HObject)s1 ) ) {
    * evaled = true;
    * put(j-1, s2);
    * put(j, s1);
    * k = j;
    * }
    * }                        // for
    * l = k + 1;
    * for(j=l;j<=r;j++) {
    * s1 = at(j-1);
    * s2 = a[j];
    * if ( ((HObject)s2).less( (HObject)s1 ) ) {
    * evaled = true;
    * put(j-1, s2);
    * put(j, s1);
    * k = j;
    * }
    * }                               // for
    * r = k - 1;
    * } while (l <= r);                 // do .. while
    * if (evaled) {
    * if (C.traceAttributeFlag) {
    * C.traceOut.append( "TRACE::ORDERLESS ");
    * C.traceOut.append( toString() + "\n" );
    * }
    * }
    * }                                   // if
    * return evaled;
    * }
    */

   /**
    * Evaluate the arguments of the function
    * The function arguments evaluation is restricted by
    * the attributes HSymbol.HOLDFIRST and HSymbol.HOLDREST
    * 
    * example: suggest the Symbol f has the attribute HSymbol.HOLDFIRST
    * f(1+1, 3*3) ==> f(1+1, 9)
    * @see HSymbol
    */
   public HFunction evalArguments(int attr)
   {
      HObject   temp;

      HFunction res = (HFunction) clone();
      boolean   isEvaled = false;

      if ((HSymbol.HOLDFIRST & attr) != HSymbol.HOLDFIRST)
      {         // evaluate first argument ?
         temp = C.ELoop(a[0]);

         if (temp != null)
         {
            res.a[0] = temp;

            isEvaled = true;
         }
      }

      if ((HSymbol.HOLDREST & attr) != HSymbol.HOLDREST)
      {         // evaluate rest of arguments ?
         for (int i = 1; i < len; i++)
         {
            temp = C.ELoop(a[i]);

            if (temp != null)
            {
               res.a[i] = temp;

               isEvaled = true;
            }
         }      // end for
      }

      if (isEvaled)
      {
         return res;
      }

      return null;
   }

   /**
    * Evaluate the function for the attributes; the function contains only 1 argument!
    * HSymbol.HOLDALL
    * HSymbol.HOLDFIRST
    * HSymbol.HOLDREST
    * HSymbol.FLAT
    * HSymbol.LISTABLE
    * HSymbol.ORDERLESS
    * 
    * @see HSymbol
    * @return <code>null</code> if the function can't be evaluated
    * <code>evaluated Object</code> if the function can be evaluated
    */
   public HFunction evaluateAttributes1()
   {
      HFunction result;
      HObject   temp;
      int       attr = symbol.getAttributes();
       /*
      if ((HSymbol.ONEIDENTITY & attr) == HSymbol.ONEIDENTITY)
      {
         return  a[0];
      }  */

      if (((HSymbol.HOLDALL & attr) != HSymbol.HOLDALL) 
          && ((HSymbol.HOLDFIRST & attr) != HSymbol.HOLDFIRST))
      {         // evaluate the arguments ?

         if ((temp = C.ELoop(a[0])) != null)
         {
            return new HFunction(symbol, temp);
            // result = (HFunction) clone();
            // result.setElementAt(temp, 0);
            // return result;
         }
      }

      if (a[0] instanceof HFunction)
      {
         HFunction      lst = (HFunction) a[0];

         if (((FLAT_EVALED & prop) != FLAT_EVALED) 
             && ((HSymbol.FLAT & attr) == HSymbol.FLAT))
         {      // associative
            if (symbol.equals(a[0].head()))
            {
              ((HFunction)a[0]).prop |= FLAT_EVALED;
              return (HFunction)a[0];
            } /*
            if ((result = flatten()) != null)
            {
               result.prop |= FLAT_EVALED;

               return result;
            }   */
         }

         if ((HSymbol.LISTABLE & attr) == HSymbol.LISTABLE)
         {      // thread over the lists
            if (a[0] instanceof HFunction && ((HFunction)a[0]).symbol==C.List)
            {
               result = new HFunction(C.List, lst.len);

               for (int j = 0; j < lst.len; j++)
               {
                  result.a[j] = new HFunction(symbol, lst.a[j]);
               }

               return result;
            }
         }
      }

      return null;
   }

   /**
    * Evaluate the function for the attributes; the function contains only 2 arguments!
    * HSymbol.HOLDALL
    * HSymbol.HOLDFIRST
    * HSymbol.HOLDREST
    * HSymbol.FLAT
    * HSymbol.LISTABLE
    * HSymbol.ORDERLESS
    * 
    * @see HSymbol
    * @return <code>null</code> if the function can't be evaluated
    * <code>evaluated Object</code> if the function can be evaluated
    */
   public HFunction evaluateAttributes2()
   {
      HFunction result;
      int       attr = symbol.getAttributes();

      if ((HSymbol.HOLDALL & attr) != HSymbol.HOLDALL)
      {         // evaluate the arguments ?
         if ((result = evalArguments(attr)) != null)
         {
            return result;
         }
      }

      if ( (a[0] instanceof HFunction) || 
      	   (a[1] instanceof HFunction))
      {
         if (((FLAT_EVALED & prop) != FLAT_EVALED) 
             && ((HSymbol.FLAT & attr) == HSymbol.FLAT))
         {      // associative
            if ( (symbol.equals( a[0].head())) ||
                 (symbol.equals( a[1].head())) )
            {
              result = new HFunction(symbol);
              if (symbol.equals(( a[0]).head()))
              {
                 result.addAll( (HFunction) a[0] );
              }
              else
              {
                 result.add( a[0]);
              }
              if (symbol.equals(( a[1]).head()))
              {
                 result.addAll( (HFunction) a[1] );
              }
              else
              {
                 result.add( a[1]);
              }
              result.prop |= FLAT_EVALED;
              return result;
            }
            /*
            if ((result = flatten()) != null)
            {
               result.prop |= FLAT_EVALED;

               return result;
            } */
         }

         if ((HSymbol.LISTABLE & attr) == HSymbol.LISTABLE)
         {      // thread over the lists
            int       listLength = 0;
            if (a[0] instanceof HFunction && ((HFunction)a[0]).symbol==C.List)
            {
              listLength = ((HFunction)a[0]).len;
            }
            if (a[1] instanceof HFunction && ((HFunction)a[1]).symbol==C.List)
            {
               if (listLength == 0)
               {   
                  listLength = ((HFunction)a[1]).len;
               }
               else
               {
                  if (listLength != ((HFunction)a[1]).len)
                  {   
                     listLength = 0; 
                  }   
               }   
            }
            if ((listLength != 0) && (result = threadList(listLength)) != null)
            {
               return result;
            }
         }
      }

      if (((ORDERLESS_EVALED & prop) != ORDERLESS_EVALED) 
          && ((HSymbol.ORDERLESS & attr) == HSymbol.ORDERLESS))
      {         // commutative
         if ( (a[1]).less(  a[0]) )
         {
            // swap the 2 arguments:
            HObject      temp = a[0];
            a[0] = a[1];
            a[1] = temp;
            prop |= ORDERLESS_EVALED;
         }
      }

      return null;
   }

   /**
    * Evaluate the function for the attributes
    * HSymbol.HOLDALL
    * HSymbol.HOLDFIRST
    * HSymbol.HOLDREST
    * HSymbol.FLAT
    * HSymbol.LISTABLE
    * HSymbol.ORDERLESS
    * 
    * @see HSymbol
    * @return <code>null</code> if the function can't be evaluated
    * <code>evaluated Object</code> if the function can be evaluated
    */
   public HFunction evaluateAttributes()
   {
      HFunction result;
      int       attr = symbol.getAttributes();

      if (len > 0)
      {                 // don't optimize this
         if ((HSymbol.HOLDALL & attr) != HSymbol.HOLDALL)
         {              // evaluate the arguments ?
            if ((result = evalArguments(attr)) != null)
            {
               return result;
            }
         }

         if (((FLAT_EVALED & prop) != FLAT_EVALED) 
             && ((HSymbol.FLAT & attr) == HSymbol.FLAT))
         {              // associative
            if ((result = flatten()) != null)
            {
               result.prop |= FLAT_EVALED;

               return result;
            }
         }

         if ((HSymbol.LISTABLE & attr) == HSymbol.LISTABLE)
         {              // thread over the lists
            int       listLength = 0;

            for (int i = 0; i < len; i++)
            {
               if (a[i] instanceof HFunction && ((HFunction)a[i]).symbol==C.List)
               {
                  if (listLength == 0)
                  {
                     listLength = ((HFunction)a[i]).len;
                  }
                  else
                  {
                     if (listLength != ((HFunction)a[i]).len)
                     {
                       listLength = 0;
                       break;    // for loop
                     }
                  }
               }
            }
            if ((listLength != 0) && (result = threadList(listLength)) != null)
            {
               return result;
            }
         }

         if (len > 1)
         {
            if (((ORDERLESS_EVALED & prop) != ORDERLESS_EVALED) 
                && ((HSymbol.ORDERLESS & attr) == HSymbol.ORDERLESS))
            {           // commutative
               sort();

               prop |= ORDERLESS_EVALED;
            }
         }
      }

      return null;
   }
   public HObject evaluateArgAttr()
   {
      if (len >= 1)
      {
         if (len == 1)
         {
            if ((HSymbol.ONEIDENTITY & symbol.getAttributes()) == HSymbol.ONEIDENTITY)
            {
              return a[0];
            }
            return evaluateAttributes1();
         }
         else
         {
            if (len == 2)
            {
               return evaluateAttributes2();
            }
            else
            {
               return evaluateAttributes();
            }
         }
      }
      return null;
   }
   public HFunction evaluateLHSAttr()
   {
      if (len >= 1)
      {
         if (len == 1)
         {
            return evaluateAttributes1();
         }
         else
         {
            if (len == 2)
            {
               return evaluateAttributes2();
            }
            else
            {
               return evaluateAttributes();
            }
         }
      }
      return null;
   }
   /**
    * Method declaration
    * 
    * 
    * @param head
    * 
    * @return
    * 
    * @see
    */
   public HObject apply(HObject head)
   {
      if (C.DEBUG)
      {         // only in debug mode
         if (head == null)
         {
            throw new HThrowException(C.HeadFormatError, C.Lambda, C.Null);
         }
      }

      if (head instanceof HSymbol)
      {
         return C.EV(new HFunction((HSymbol) head, (HArrayList) this));
      }

      if ((head instanceof HFunction) && ((HFunction) head).isPureFunction())
      {
         HFunction      l = (HFunction) ((HFunction) head).clone();
         l.addAll( this );
         /*
         for (int i = 0; i < len; i++)
         {
            l.add( a[i] );
         }*/

         return C.EV(l);
      }

      throw new HThrowException(C.HeadFormatError, C.Lambda, head);
   }

   /**
    * Evaluate the function
    * The function arguments evaluation is restricted by
    * the attributes of the header symbol
    * These attributes are:
    * HSymbol.ONEIDENTITY
    * HSymbol.HOLDALL
    * HSymbol.HOLDFIRST
    * HSymbol.HOLDREST
    * HSymbol.FLAT
    * HSymbol.LISTABLE
    * HSymbol.ORDERLESS
    * 
    * @see HSymbol
    * @return <code>null</code> if the function can't be evaluated
    * <code>evaluated Object</code> if the function can be evaluated
    */
   public HObject evaluate()
   {
      return symbol.evaluateFunction(this);
   }

   /**
    * Replace the (sub-)objects with the rule in HRuleSet
    * 
    * @return     null if no replacement occured;
    * a new object otherwise
    */
   public HObject replaceAll(HRuleSet rset)
   {
      boolean   evaled = false;
      HObject   result;
      HFunction temp = (HFunction) clone();

      for (int i = 0; i < len; i++)
      {
         result = rset.getDownRule( a[i]);

         if (result != null)
         {
            temp.set(i, result);

            evaled = true;
         }
         else
         {
            if (a[i] instanceof HFunction)
            {
               result = ((HFunction) a[i]).replaceAll(rset);

               if (result != null)
               {
                  temp.set(i, result);

                  evaled = true;
               }
            }
         }
      }

      if (evaled)
      {
         return temp;
      }

      return null;
   }

   /**
    * Substitute the HObject <code>obj</code> in <code>this</code>
    * with the HObject <code>subst</code>
    * 
    * @param obj         an object to compare with
    * @param subst       the substitution object
    */
   public HObject substitute(HObject obj, HObject subst)
   {
      boolean   evaled = false;
      HFunction temp = (HFunction) clone();

      if (equals(obj))
      {
         return subst;
      }

      if (symbol.equals(obj) && (subst instanceof HSymbol))
      {
         temp.symbol = (HSymbol) subst;
         evaled = true;
      }

      HObject   tempi;

      for (int i = 0; i < len; i++)
      {
         tempi = ( a[i]).substitute(obj, subst);

         if (tempi !=  a[i])
         {
            temp.set(i, tempi);

            evaled = true;
         }
      }

      if (evaled)
      {
         return temp;
      }

      return this;
   }

   public HObject substituteArg(ObjectEvaluator eval)
   {
      boolean   evaled = false;
      HFunction temp = (HFunction) clone();
      HObject   tobj;

      for (int i = 0; i < len; i++)
      {
        if ( ( tobj = eval.evaluate( a[i] ) ) != null)
        {
          temp.set(i, tobj);
          evaled = true;
        }
      }
      if (evaled)
      {
         return temp;
      }
      return null;
   }
   /**
    * Substitute all patterns in <code>this</code> with their values
    */
   public HObject substitutePattern()
   {
      boolean   evaled = false;
      HFunction temp = (HFunction) clone();

      if ((!symbol.patternValue.isEmpty()) 
          && (symbol.patternValue.peek() instanceof HSymbol))
      {
         temp.symbol = (HSymbol) symbol.patternValue.peek();
         evaled = true;
      }

      HObject   tempi;

      for (int i = 0; i < len; i++)
      {
         tempi = ( a[i]).substitutePattern();

         if (tempi != null && tempi !=  a[i])
         {

            // temp.setElementAt(((HObject) a[i]).substitutePattern(), i);
            temp.set(i, tempi);

            evaled = true;
         }
      }

      if (evaled)
      {
         return temp;
      }

      return this;
   }

   /*
    * public boolean oldMatches(HObject obj) {
    * if (obj instanceof HFunction) {
    * HFunction temp = (HFunction) obj;
    * if ( symbol.unequals( temp.symbol ) ) {
    * return false;
    * }
    * 
    * if ( (len < temp.len ) &&
    * ( len >= 1) ) {
    * if ( (HSymbol.FLAT & temp.head().getAttributes()) == HSymbol.FLAT) { // associative
    * for (int i=0;i<len-1;i++) {
    * if ( ! ((HObject)a[i]).matches( (HObject) temp.a[i] ) ) {
    * return false;
    * }
    * }
    * HFunction mt = new HFunction( temp, symbol, len-1, temp.len, true );
    * if ( elementAt(len-1).matches( mt ) ) {
    * return true;
    * }
    * return false;
    * }
    * else {
    * return false;
    * }
    * }
    * if ( len != temp.len ) {
    * return false;
    * }
    * for (int i=0;i<len;i++) {
    * if ( ! ((HObject)a[i]).matches( (HObject) temp.a[i] ) ) {
    * return false;
    * }
    * }
    * return true;
    * }
    * return false;
    * }
    */

   /**
    * Method Declaration.
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public boolean matches(HObject obj)
   {
      if (C.DEBUG)
      {     /*
         if (C.traceFlag && C.traceExtendedFlag)
         {
            C.traceOut.append("TRACE(HFunction::matches)::");
            C.traceOut.append(toString());
            C.traceOut.append(" {");
            C.traceOut.append(new Integer(prop).toString());
            C.traceOut.append("} <-> ");
            C.traceOut.append(obj.toString());
         }    */
      }

      if (obj instanceof HFunction)
      {
         if ((prop & MATCHEQUAL) == MATCHEQUAL)
         {
            if (C.DEBUG)
            {          /*
               if (C.traceFlag && C.traceExtendedFlag)
               {
                  C.traceOut.append(" MATCHEQUAL\n");
               }         */
            }

            return equals(obj);
         }

         HFunction      temp = (HFunction) obj;

         if (symbol.equals(temp.symbol) && len == temp.len)
         {
            for (int i = 0; i < len; i++)
            {
               if (!( a[i]).matches( temp.a[i]))
               {
                  if (C.DEBUG)
                  {          /*
                     if (C.traceFlag && C.traceExtendedFlag)
                     {
                        C.traceOut.append(" FALSE at i\n");
                     }         */
                  }

                  return false;
               }
            }

            if (C.DEBUG)
            {          /*
               if (C.traceFlag && C.traceExtendedFlag)
               {
                  C.traceOut.append(" TRUE\n");
               }         */
            }

            return true;
         }

         if (C.DEBUG)
         {              /*
            if (C.traceFlag && C.traceExtendedFlag)
            {
               C.traceOut.append(" {");
               C.traceOut.append(new Integer(this.len).toString());
               C.traceOut.append(",");
               C.traceOut.append(new Integer(temp.len).toString());
               C.traceOut.append("} FALSE end\n");
            }             */
         }
      }

      return false;
   }

   /**
    * Matches all simple patterns of the function
    * and return "difficult" pattern-matching Pairs on the Stack
    */

   /*
    * public boolean patternMatches(HObject obj, Stack stck) {
    * if (C.DEBUG) {
    * if (C.traceFlag && C.traceExtendedFlag) {
    * C.traceOut.append("TRACE(try patternMatches)::");
    * C.traceOut.append(toString());
    * C.traceOut.append(" {");
    * C.traceOut.append(new Integer(prop).toString());
    * C.traceOut.append("} <-> ");
    * C.traceOut.append(obj.toString());
    * C.traceOut.append("}\n");
    * }
    * }
    * if ((len == 1)
    * && ((HSymbol.ONEIDENTITY & symbol.getAttributes())
    * == HSymbol.ONEIDENTITY)) {
    * if (((HObject) a[0]).matches(obj)) {
    * return true;
    * }
    * }
    * if ((prop & MATCHEQUAL) == MATCHEQUAL) {
    * return equals(obj);
    * }
    * if ((prop & MATCHSEQUENCE) == MATCHSEQUENCE) {
    * return matches(obj);
    * }
    * if (obj instanceof HFunction) {
    * HFunction	temp = (HFunction) obj;
    * if (symbol.unequals(temp.symbol)) {
    * return false;
    * }
    * if ((len < temp.len) && (len > 0)) {
    * if ((HSymbol.FLAT & temp.head().getAttributes())
    * == HSymbol.FLAT) {
    * stck.push(new Pair(this, temp));
    * return true;
    * } else {
    * return false;
    * }
    * }
    * if (len != temp.len) {
    * return false;
    * }
    * if ((HSymbol.ORDERLESS & temp.head().getAttributes())
    * == HSymbol.ORDERLESS) {
    * stck.push(new Pair(this, temp));
    * return true;
    * }
    * for (int i = 0; i < len; i++) {
    * if (a[i] instanceof HFunction) {
    * if (!((HFunction) a[i]).patternMatches((HObject) temp.a[i],
    * stck)) {
    * return false;
    * }
    * } else {
    * if (!((HObject) a[i]).matches((HObject) temp.a[i])) {
    * return false;
    * }
    * }
    * }
    * return true;
    * }
    * return false;
    * }
    */

   /**
    * Determines, if <code>this</code> can be matched with HObject.equals()
    * in the left-hand-side of a pattern-matching rule
    */
   public int isEqualRule()
   {

      // HArrayList pats = new HArrayList();

      int       patFlag = 0;
      int       temp;

      for (int i = 0; i < len; i++)
      {
         if (( a[i]) instanceof HPattern)
         {
            if (patFlag == 0)
            {
               patFlag = 1;
            }
         }
         else
         {
            if (( a[i]) instanceof HFunction)
            {
               temp = ((HFunction) a[i]).isEqualRule();

               if (temp > patFlag)
               {
                  patFlag = temp;
               }
            }
         }
      }

      int       attr = symbol.getAttributes();

      if (patFlag > 0 
          && ((attr & HSymbol.FLATORDERLESS) != HSymbol.NOATTRIBUTE))
      {
         prop &= NOPROPERTY;

         return 2;
      }

      switch (patFlag)
      {

         case 0:
            setProperties(MATCHEQUAL);

            return 0;

         case 1:
            setProperties(MATCHSEQUENCE);

            return 1;

         default:
            prop &= NOPROPERTY;

            return 2;
      }
   }

   /**
    * Returns a specific integer value for each class in the
    * math-objects hierarchy
    */
   final public int hierarchy()
   {
      return HObject.FUNCTIONID;  // 64
   }

   /**
    * Method Declaration.
    * 
    * 
    * @param os
    * 
    * @exception IOException
    * 
    * @see
    */   /*
   public void writeTo(DataOutputStream os) throws IOException
   {
      int       sz = len;
      int       indx = -1;

      if (sz <= 5)
      {
         for (int i = 0; i < C.MAXIOSYMBOLS; i++)
         {
            if (symbol.equals(C.IOSymbolTable[i]))
            {
               indx = i;

               break;
            }
         }

         if (indx != -1)
         {
            os.writeByte((byte) (80 + sz));
            os.writeByte((byte) (indx));

            for (int i = 0; i < sz; i++)
            {
               ( a[i]).writeTo(os);
            }
         }
         else
         {
            os.writeByte((byte) (HObject.FUNCTIONFILEID - sz - 1));
            symbol.writeTo(os);

            for (int i = 0; i < sz; i++)
            {
               ( a[i]).writeTo(os);
            }
         }
      }
      else
      {
         os.writeByte(HObject.FUNCTIONFILEID);
         os.writeInt(sz);
         symbol.writeTo(os);

         for (int i = 0; i < sz; i++)
         {
            ( a[i]).writeTo(os);
         }
      }
   }       */

   /**
    * Method Declaration.
    *
    *
    * @return
    * 
    * @see
    */
   final public HSymbol head()
   {
      return symbol;
   }

   /*
    * public boolean isInstance(Class cls) {
    * for (int i=0;i<len;i++) {
    * if (! cls.isInstance( a[i] ) ) {
    * return false;
    * }
    * }
    * return true;
    * }
    */

   /**
    * All arguments of this are instances of HDouble or HDoubleComplex
    * 
    * 
    * @return
    * 
    * @see
    */
   public boolean isNumericNumber()
   {
      for (int i = 0; i < len; i++)
      {
         if (a[i] instanceof HDouble || a[i] instanceof HDoubleComplex)
         {
            continue;
         }

         return false;
      }

      return true;
   }

   /**
    * Method declaration
    * 
    * 
    * @return
    * 
    * @see
    */
   final public boolean isPureFunction()
   {
      return (symbol.equals(C.Lambda) && len == 2);
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   final public boolean isList()
   {
      return (symbol.equals(C.List));
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public boolean isNumber()
   {
      return false;
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public boolean isRational()
   {
      return false;
   }

   /**
    * Returns the index if this contains the obj as a member; otherwise -1
    */
   public int findMember(HObject obj)
   {
      for (int i = 0; i < len; i++)
      {
         if (( a[i]).equals(obj))
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Returns true if this contains the HObject obj
    */
   public boolean isMember(HObject obj)
   {
      if (equals(obj))
      {
         return true;
      }

      for (int i = 0; i < len; i++)
      {
         if ((a[i]).isMember(obj))
         {
            return true;
         }
      }

      return false;
   }

   /**
    * Return <code>true</code> if the argument equals the head of <code>this</code> or any subobject head
    * 
    * @see HFunction
    */
   public boolean isHeadMember(HSymbol sym)
   {
      if (symbol.equals(sym))
      {
         return true;
      }

      for (int i = 0; i < len; i++)
      {
         if (( a[i]).isHeadMember(sym))
         {
            return true;
         }
      }

      return false;
   }

   /**
    * Method declaration
    * 
    * 
    * @param obj
    * 
    * @return
    * 
    * @see
    */
   public HFunction map(HObject obj)
   {
      HFunction f = new HFunction(head(), len);

      for (int i = 0; i < len; i++)
      {
         f.set(i, new HFunction(C.Null, a[i]).apply(obj));
      }

      return f;
   }


   /**
    * Method declaration
    * 
    * 
    * @param f
    * @param obj
    * @param level
    * @param range
    * 
    * @return
    * 
    * @see
    */
   public int map(HFunction f, HObject obj, int level, HLevel range)
   {

      // HFunction h = null;
      HObject   temp;
      int       negLevel = 0;
      int       elemLevel = 0;

      level++;

      for (int i = 0; i < len; i++)
      {
         temp = a[i];

         if (a[i] instanceof HFunction)
         {
            HFunction   g = new HFunction((a[i]).head());

            elemLevel = ((HFunction) a[i]).map(g, obj, level, range);
            temp = g;
         }
         else
         {
            elemLevel = -1;
         }

         if (elemLevel < negLevel)
         {
            negLevel = elemLevel;
         }

         if (range.includes(elemLevel, level))
         {

            // h = new HFunction(C.Null, temp);
            f.add(new HFunction(C.Null,  temp).apply(obj));
         }
         else
         {
            f.add(temp);
         }
      }

      return --negLevel;
   }

   /**
    * Method declaration
    * 
    * 
    * @param arr
    * @param level
    * @param range
    * 
    * @return
    * 
    * @see
    */
   public int level(HArrayList arr, int level, HLevel range)
   {
      int       negLevel = 0;
      int       elemLevel = 0;

      level++;

      for (int i = 0; i < len; i++)
      {
         if (a[i] instanceof HFunction)
         {
            elemLevel = ((HFunction) a[i]).level(arr, level, range);
         }
         else
         {
            elemLevel = -1;
         }

         if (elemLevel < negLevel)
         {
            negLevel = elemLevel;
         }

         if (range.includes(elemLevel, level))
         {
            arr.add(a[i]);
         }
      }

      return --negLevel;
   }

   /**
    * Method declaration
    * 
    * 
    * @return
    * 
    * @see
    */
   public int depth()
   {
      int       lev = 0;
      int       elemLevel = 0;

      for (int i = 0; i < len; i++)
      {
         if (a[i] instanceof HFunction)
         {
            elemLevel = ((HFunction) a[i]).depth();
         }
         else
         {
            elemLevel = 1;
         }

         if (elemLevel > lev)
         {
            lev = elemLevel;
         }
      }

      return ++lev;
   }

   /**
    * Returns all (sub-)expressions in an HArrayList-list, which are an instance of
    * the specified Class
    */
   public void getInstanceList(HArrayList lst, Class cls)
   {
      if (cls.isInstance(this))
      {
         lst.add(this);
      }

      for (int i = 0; i < len; i++)
      {
         if (a[i] instanceof HFunction)
         {
            ((HFunction) a[i]).getInstanceList(lst, cls);
         }
         else
         {
            if (cls.isInstance(a[i]))
            {
               lst.add(a[i]);
            }
         }
      }
   }

   /**
    * Method Declaration.
    * 
    * 
    * @return
    * 
    * @see
    */
   public int precedence()
   {
      return symbol.precedence();
   }

}