/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package com.hartmath.expression;

import java.lang.*;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Interface that represents a math object:
 * e.g integer, fraction, complex, string, symbol, pattern, function,...
 * 
 * 
 * @see HDouble
 * @see HDoubleComplex
 * @see HInteger
 * @see HFraction
 * @see HComplex
 * @see HString
 * @see HSymbol
 * @see HPattern
 * @see HFunction
 * 
 * @author Klaus Hartlage  <A HREF="mailto:khartlage@t-online.de">khartlage@t-online.de</A>
 */
public interface HObject
{
   // ID's for the hierarchy function.
   // define a bit for every type:
   static public final int DOUBLEID        = 1;
   static public final int DOUBLECOMPLEXID = 2;
   static public final int INTEGERID       = 4;
   static public final int FRACTIONID      = 8;
   static public final int COMPLEXID       = 16;
   static public final int STRINGID        = 32;
   static public final int SYMBOLID        = 64;
   static public final int FUNCTIONID      = 128;
   static public final int PATTERNID       = 256;
   // ID's for init.hm file
   static public final byte DOUBLEFILEID        = 2;
   static public final byte DOUBLECOMPLEXFILEID = 3;
   static public final byte INTEGERFILEID       = 4;
   static public final byte FRACTIONFILEID      = 5;
   static public final byte COMPLEXFILEID       = 6;
   static public final byte STRINGFILEID        = 16;
   static public final byte SYMBOLFILEID        = 32;
   static public final byte FUNCTIONFILEID      = 64;
   static public final byte PATTERNFILEID       = 65;
   /**
    * Returns true if <code>this</code> is not equal to <code>obj</code>
    * @param obj         an object to compare with
    */
   abstract public boolean unequals(Object obj);

   /**
    * Returns true if <code>this</code> is less to <code>obj</code>
    * @param obj         an object to compare with
    */
   abstract public boolean less(Object obj);

   /**
    * Returns true if <code>this</code> is greater to <code>obj</code>
    * @param obj         an object to compare with
    */
   abstract public boolean greater(Object obj);

   /**
    * Evaluate this.
    * 
    * @return a defined math object otherwise return
    * <code>null</code> if the object could not be evaluated
    */
   abstract public HObject evaluate();

   /**
    * Returns true if <code>this</code> (pattern-)matches the object
    * <code>obj</code>
    * @param obj         an object to compare with
    */
   abstract public boolean matches(HObject obj);

   /**
    * Substitute the HObject obj in this with the HObject subst
    * 
    * @param obj         an object to compare with
    * @param subst       the substitution for the object
    */
   abstract public HObject substitute(HObject obj, HObject subst);

   /**
    * Substitute all HPattern-objects in this with the HObject subst
    * 
    * @param subst       the substitution for the object
    */
   abstract public HObject substitutePattern();

   abstract public void toStringBuffer(StringBuffer buf);
   /**
    * Returns an specific integer value for each class in the
    * HObject class-hierarchy
    */
   abstract public int hierarchy();

   /**
    * Write a persistant object
    */
  // abstract public void writeTo(DataOutputStream os) throws IOException;

   /**
    * Returns the associated header-symbol of this HObject
    * e.g for a HInteger object the symbol C.Integer will be returned
    * and for a HFunction the special header-symbol will be retuned
    * 
    * @see HFunction
    */
   abstract public HSymbol head();

   /**
    * Method declaration
    *
    *
    * @param head
    *
    * @return
    *
    * @see
    */
   abstract public HObject apply(HObject head);

   /**
    * Returns true if this is an instance of HFunction
    * with the header symbol C.List
    * 
    * @ see C
    */
   abstract public boolean isList();

   /**
    * Returns true if this is a number
    * e.g. HInteger, HFraction, HComplex, HDouble or HDoubleComplex
    */
   abstract public boolean isNumber();

   /**
    * Returns true if this is a rational number
    * e.g. HInteger, HFraction
    */
   abstract public boolean isRational();

   /**
    * Returns true if this contains the HObject obj
    */
   abstract public boolean isMember(HObject obj);

   /**
    * Returns true if this contains the HSymbol sym as a header
    */
   abstract public boolean isHeadMember(HSymbol sym);

   /**
    * Operator precedence of the mathematical object
    * 
    * @return the precedence of the object
    * (important for a function which should be displayed in operator notation
    */
   abstract public int precedence();
}



