package com.hartmath.JSci.maths.statistics;

import com.hartmath.expression.*;
import com.hartmath.util.*;
import com.hartmath.lib.*;

/**
* The PoissonDistribution class provides an object for encapsulating Poisson distributions.
* @version 0.2
* @author Mark Hale
*/
public final class PoissonDistribution extends ProbabilityDistribution {
        private double lambda;

        /**
        * Constructs a Poisson distribution.
        * @param interval the interval
        */
        public PoissonDistribution(double interval) {
                if(interval<=0.0)
                {
                  throw new HThrowException(C.OutOfRangeError,C.OutOfRangeError,
                          new HString("The interval should be (strictly) positive.") );
                }
                lambda=interval;
        }
        /**
        * Returns the natural logarithm of the factorial.
        * (Wrapper for the log gamma function).
        * @see SpecialMath#logGamma
        * @param x a double
        */
        public static double logFactorial(double x) {
                return HDoubleSpecial.logGamma(x+1.0);
        }
        /**
        * Returns the interval.
        */
        public double getInterval() {
                return lambda;
        }
        /**
        * Probability density function of a Poisson distribution.
        * @param X should be integer-valued
        * @return the probability that a stochastic variable x has the value X, i.e. P(x=X)
        */
        public double probability(double X) {
                checkRange(X,0.0,Double.MAX_VALUE);
                return Math.exp(X*Math.log(lambda)-lambda-logFactorial(X));
        }
        /**
        * Cumulative Poisson distribution function.
        * @param X should be integer-valued
	* @return the probability that a stochastic variable x is less then X, i.e. P(x&lt;X)
        */
        public double cumulative(double X) {
                checkRange(X,0.0,Double.MAX_VALUE);
                double sum=0.0;
                for(double i=0.0;i<=X;i++)
                        sum+=probability(i);
                return sum;
        }
        /**
	* Inverse of the cumulative Poisson distribution function.
        * @return the value X for which P(x&lt;X)
        */
        public double inverse(double probability) {
                checkRange(probability);
                return Math.floor(findRoot(probability,lambda,0.0,Double.MAX_VALUE));
        }
}


